
from __future__ import absolute_import

import sys
import time

from ..wrapper import Type, Value
from .common import alarm, timeStamp, NTBase
from .scalar import _metaHelper, ntwrappercommon

if sys.version_info >= (3, 0):
    unicode = str

class ntenum(ntwrappercommon, int):
    """

    * .severity
    * .status
    * .timestamp - Seconds since 1 Jan 1970 UTC as a float
    * .raw_stamp - A tuple (seconds, nanoseconds)
    * .raw - The underlying :py:class:`p4p.Value`.
    """

    def _store(self, value):
        ntwrappercommon._store(self, value)
        self.choice = None

        return self

    def __str__(self):
        return self.choice or int.__str__(self)

    def __repr__(self):
        return '%s(%d, %s)'%(self.__class__.__name__, int(self), self.choice)

    # TODO: compare with str

class NTEnum(NTBase):
    """Describes a string selected from among a list of possible choices.  Stored internally as an integer
    """
    @staticmethod
    def buildType(extra=[], display=False, control=False, valueAlarm=False):
        F = [
            ('value', ('S', 'enum_t', [
                ('index', 'i'),
                ('choices', 'as'),
            ])),
            ('alarm', alarm),
            ('timeStamp', timeStamp),
        ]
        # TODO: different metadata options
        _metaHelper(F, 'i', display=display, control=control, valueAlarm=valueAlarm)
        F.extend(extra)
        return Type(id="epics:nt/NTEnum:1.0", spec=F)

    def __init__(self, **kws):
        self.type = self.buildType(**kws)
        # cached choices list.
        # picked up during unwrap
        self._choices = []

    def wrap(self, value, choices=None, **kws):
        """Pack python value into Value
        """
        V = self.type()
        if choices is not None:
            V['value.choices'] = choices

        if isinstance(value, dict):
            # assume dict of index and choices list
            V.value = value
            self._choices = V['value.choices']
        else:
            # index or string
            self.assign(V, value)
        return self._annotate(V, **kws)

    def unwrap(self, value):
        """Unpack a Value into an augmented python type (selected from the 'value' field)
        """
        if value.changed('value.choices'):
            self._choices = value['value.choices']

        idx = value['value.index']
        ret = ntenum(idx)._store(value)
        try:
            ret.choice = self._choices[idx]
        except IndexError:
            pass # leave it as None
        return ret

    def assign(self, V, py):
        """Store python value in Value
        """
        if isinstance(py, (bytes, unicode)):
            for i,C in enumerate(V['value.choices'] or self._choices):
                if py==C:
                    V['value.index'] = i
                    return
            # attempt to parse as integer
            py = int(py, 0)
        else:
            # attempt to cast as integer
            py = int(py)

        V['value.index'] = py




import logging

from functools import partial

import asyncio

from .raw import SharedPV as _SharedPV, Handler
from ..client.thread import RemoteError
from ..client.asyncio import get_running_loop, create_task, all_tasks

__all__ = (
    'SharedPV',
        'Handler',
)

_log = logging.getLogger(__name__)

def _log_err(V):
    if isinstance(V, Exception):
        _log.error("Unhandled from SharedPV handler: %s", V)
        # TODO: figure out how to show stack trace...
        # until then, propagate in the hope that someone else will
    return V


def _handle(pv, op, M, args): # callback in asyncio loop
    try:
        _log.debug('SERVER HANDLE %s %s %r', op, M, args)
        maybeco = M(*args)
        if asyncio.iscoroutine(maybeco):
            _log.debug('SERVER SCHEDULE %s', maybeco)
            task = create_task(maybeco)

            # we have no good place to join async put()/rpc() handler results
            # other than SharedPV.close(sync=True) which is both optional,
            # and potentially far in the future.  So we log and otherwise
            # discard the result.
            task.add_done_callback(_log_err)
            task._SharedPV = pv # mark so _wait_closed() can distinguish
        return # caller is responsible for op.done()
    except RemoteError as e:
        err = e
    except Exception as e:
        _log.exception("Unexpected")
        err = e
    if op is not None:
        op.done(error=str(err))


class SharedPV(_SharedPV):

    def __init__(self, handler=None, **kws):
        self.loop = get_running_loop()
        _SharedPV.__init__(self, handler=handler, **kws)
        self._disconnected = asyncio.Event()
        self._disconnected.set()

    def _exec(self, op, M, *args):
        # note than M may be _onFirstConnect or _onLastDisconnect
        self.loop.call_soon_threadsafe(partial(_handle, self, op, M, args))

    def _onFirstConnect(self, _junk):
        self._disconnected.clear()

    def _onLastDisconnect(self, _junk):
        self._disconnected.set()

    async def _wait_closed(self):
        """Wait until any in-progress asynchronous put()/rpc() handler tasks have completed.
        """
        _log.debug("Synchronizing %r", self)

        def _peak_done(F, V):
            F.set_result(V)
            return V

        Ts = []
        for t in all_tasks():
            if getattr(t, '_SharedPV', None) is not self:
                continue
            F = asyncio.Future()
            t.add_done_callback(partial(_peak_done, F))
            Ts.append(F)

        await asyncio.gather(*Ts, return_exceptions=True)
        # ignore any returned exceptions as they have already been logged

        _log.debug("Synchronized %r", self)

        # wait for Disconnect notification as well
        await self._disconnected.wait()

    def close(self, destroy=False, sync=False):
        """Close PV, disconnecting any clients.

        :param bool destroy: Indicate "permanent" closure.  Current clients will not see subsequent open().
        :param bool sync: When block until any pending onLastDisconnect() is delivered (timeout applies).
        :param float timeout: Applies only when sync=True.  None for no timeout, otherwise a non-negative floating point value.

        close() with destory=True or sync=True will not prevent clients from re-connecting.
        New clients may prevent sync=True from succeeding.
        Prevent reconnection by __first__ stopping the Server, removing with :py:method:`StaticProvider.remove()`,
        or preventing a :py:class:`DynamicProvider` from making new channels to this SharedPV.
        """
        _SharedPV.close(self, destroy)
        if sync:
            return self._wait_closed()



import warnings
from .lex import tokens, ACFError

start = 'asconfig'

def p_item(p):
    """asconfig_item : uag
                     | hag
                     | asg
       rule_item : uag_ref
                 | hag_ref
                 | calc_head
    """
    p[0] = p[1]

def p_top_items(p):
    """uag : UAG  '(' STRING ')' string_body
       hag : HAG  '(' STRING ')' string_body
       asg : ASG  '(' STRING ')' asg_body
    """
    p[0] = (p[1], p[3], p[5])

def p_list_append(p):
    """asconfig    : asconfig        asconfig_item
       string_list : string_list ',' STRING
       asg_list    : asg_list        asg_item
       rule_list   : rule_list       rule_item
    """
    #p[0] = p[1] + [p[len(p)-1]]
    p[1].append(p[len(p)-1])
    p[0] = p[1]

def p_list_one(p):
    """asconfig    : asconfig_item
       string_list : STRING
       asg_list    : asg_item
       rule_list   : rule_item
    """
    p[0] = [p[1]]

def p_head(p):
    """uag_ref  : UAG  '(' string_list ')'
       hag_ref  : HAG  '(' string_list ')'
       calc_head : CALC '(' STRING ')'
    """
    p[0] = (p[1], p[3])

def p_body_empty(p):
    """string_body :
       asg_body    :
       rule_body   :
    """
    p[0] = None

def p_body(p):
    """string_body : '{' string_list '}'
       asg_body    : '{' asg_list    '}'
       rule_body   : '{' rule_list   '}'
    """
    p[0] = p[2]

def p_asg_inp(p):
    """asg_item : INP '(' STRING ')'
    """
    p[0] = ('INP', p[1][-1:], p[3])

def p_asg_rule(p):
    """asg_item : RULE '(' INTEGER ',' STRING trap ')' rule_body
    """
    p[0] = ('RULE', p[3], p[5], p[6], p[8])

def p_trap_empty(p):
    """trap : 
    """
    p[0] = False

def p_trap(p):
    """trap : ',' STRING
    """
    if p[2] not in ('TRAPWRITE', 'NOTRAPWRITE'):
        warnings.warn("trap spec. must be 'TRAPWRITE' or 'NOTRAPWRITE'")
        p[0] = False
    else:
        p[0] = p[2]=='TRAPWRITE'

def p_error(p):
    raise ACFError("Syntax error on line %d at or before '%s'"%(p.lineno, p.value))

def parse(acf, debug=False):
    from ply import yacc, lex
    from . import lex as _lex
    lex = lex.lex(module=_lex)
    parser = yacc.yacc(debug=debug,write_tables=False)
    return parser.parse(acf, lex, debug=debug)

if __name__=='__main__':
    import sys
    from pprint import pprint
    with open(sys.argv[1],'r') as F:
        data = F.read()
    pprint(parse(data, debug=True))



# exec()d by setup.py and gha-set-pre.py

class Version(object):
    """P4P Version number
    
    Comparible with version string, or a tuple of integers with len() <= 5.
    """
    def __init__(self, s):
        self._parts = self.__parse(s)

    @staticmethod
    def __parse(s):
        import re
        M = re.match(r'^(\d+)\.(\d+)\.(\d+)(?:([ab])(\d+))?$', s)
        if M is None:
            raise ValueError('Invalid version %r'%s)
        x,y,z,abr,n = M.groups()
        abr = {
            None:0,
            'a':-2,
            'b':-1,
        }[abr]
        assert abr is not None or n=='0', s
        return (int(x),int(y),int(z), abr, int(n or '0'))

    @property
    def is_release(self):
        return self._parts[3]==0

    def __repr__(self):
        s = str(self)
        return 'Version(%r)'%s

    def __str__(self):
        x,y,z,abr,n = self._parts
        ver = '%d.%d.%d'%(x,y,z)
        abr = {0:'', -1:'b', -2:'a'}[abr]
        if abr:
            ver = ver + '%s%d'%(abr, n)
        return ver

    def _cmp(self, o):
        if isinstance(o, str):
            o = self.__parse(o)
        o = tuple(o)
        if len(o) < 5:
            o = o + (0,)*(5 - len(o))

        elif len(o) > 5:
            raise ValueError("must compare with iterable len() <= 5")

        for l,r in zip(self._parts, o):
            if l < r:
                return -1
            elif l > r:
                return 1
        return 0

    def __lt__(self, o):
        return self._cmp(o)<0
    def __le__(self, o):
        return self._cmp(o)<=0
    def __eq__(self, o):
        return self._cmp(o)==0
    def __ge__(self, o):
        return self._cmp(o)>=0
    def __gt__(self, o):
        return self._cmp(o)>0

version = Version('4.2.1a3')



from __future__ import print_function

import unittest
import sys
import random
import weakref
import gc
import threading

from ..server import Server, installProvider, removeProvider, DynamicProvider, StaticProvider
from ..client.thread import Context
from .utils import RefTestCase


def checkweak(O):
    o = O()
    if o is not None:
        print('Live object', id(o), type(o), sys.getrefcount(o), gc.get_referrers(o))
    return o


class TestDummyProvider(RefTestCase):
    # will fail if anything is done to it.

    class Dummy(object):
        name = "foo"

    def test_install(self):
        "Install and remove provider"
        D = self.Dummy()
        d = weakref.ref(D)
        P = DynamicProvider(D.name, D)
        p = weakref.ref(P)
        installProvider("foo", P)
        del D
        del P
        removeProvider("foo")
        gc.collect()
        self.assertIsNone(d())
        self.assertIsNone(p())

    def test_server(self):
        D = self.Dummy()
        d = weakref.ref(D)
        P = DynamicProvider(D.name, D)
        p = weakref.ref(P)
        installProvider("foo", P)
        try:
            with Server(providers=["foo"]) as S:
                s = weakref.ref(S)
        finally:
            removeProvider("foo")
        del D
        del P
        del S
        gc.collect()
        self.assertIsNone(checkweak(d))
        self.assertIsNone(p())
        self.assertIsNone(checkweak(s))

    def test_server_direct(self):
        D = self.Dummy()
        d = weakref.ref(D)
        P = DynamicProvider(D.name, D)
        p = weakref.ref(P)
        with Server(providers=[P]) as S:
            s = weakref.ref(S)
        del D
        del P
        del S
        gc.collect()
        self.assertIsNone(checkweak(d))
        self.assertIsNone(p())
        self.assertIsNone(checkweak(s))

    def test_client(self):
        D = self.Dummy()
        d = weakref.ref(D)
        P = DynamicProvider(D.name, D)
        p = weakref.ref(P)
        installProvider("foo", P)
        del D
        gc.collect()
        self.assertIsNotNone(d())

        try:
            with Context('server:foo'):
                removeProvider("foo")
                # Our DynamicProvider will not longer be found by new Contexts
                # however, it remains active so long as 'P' is active

            del P
            gc.collect()
            self.assertIsNone(d())
            self.assertIsNone(p())
        finally:
            removeProvider("foo")

class TestServerConf(RefTestCase):
    def test_bad_iface(self):
        P = StaticProvider('x')
        with self.assertRaisesRegex(RuntimeError, "invalid"):
            S = Server(providers=[P], useenv=False, conf={
                'EPICS_PVAS_INTF_ADDR_LIST':'invalid.host.name.',
                'EPICS_PVAS_BROADCAST_PORT':'0',
                'EPICS_PVAS_SERVER_PORT':'0',
            })




import sys
import logging
import inspect
from functools import wraps, partial
_log = logging.getLogger(__name__)

from .wrapper import Value, Type
from .nt import NTURI
from .client.raw import RemoteError
from .server import DynamicProvider
from .server.raw import SharedPV
from .util import ThreadedWorkQueue, WorkQueue, Full, Empty

__all__ = [
    'rpc',
    'rpccall',
    'rpcproxy',
    'RemoteError',
    'WorkQueue',
    'NTURIDispatcher',
]


def rpc(rtype=None):
    """Decorator marks a method for export.

    :param type: Specifies which :py:class:`Type` this method will return.

    The return type (rtype) must be one of:

    - An instance of :py:class:`p4p.Type`
    - None, in which case the method must return a :py:class:`p4p.Value`
    - One of the NT helper classes (eg :py:class:`p4p.nt.NTScalar`).
    - A list or tuple used to construct a :py:class:`p4p.Type`.

    Exported methods raise an :py:class:`Exception` to indicate an error to the remote caller.
    :py:class:`RemoteError` may be raised to send a specific message describing the error condition.

    >>> class Example(object):
        @rpc(NTScalar.buildType('d'))
        def add(self, lhs, rhs):
            return {'value':float(lhs)+flost(rhs)}
    """
    wrap = None
    if rtype is None or isinstance(rtype, Type):
        pass
    elif isinstance(type, (list, tuple)):
        rtype = Type(rtype)
    elif hasattr(rtype, 'type'):  # eg. one of the NT* helper classes
        wrap = rtype.wrap
        rtype = rtype.type
    else:
        raise TypeError("Not supported")

    def wrapper(fn):
        if wrap is not None:
            orig = fn

            @wraps(orig)
            def wrapper2(*args, **kws):
                return wrap(orig(*args, **kws))
            fn = wrapper2

        fn._reply_Type = rtype
        return fn
    return wrapper


def rpccall(pvname, request=None, rtype=None):
    """Decorator marks a client proxy method.

    :param str pvname: The PV name, which will be formated using the 'format' argument of the proxy class constructor.
    :param request: A pvRequest string or :py:class:`p4p.Value` passed to eg. :py:meth:`p4p.client.thread.Context.rpc`.

    The method to be decorated must have all keyword arguments,
    where the keywords are type code strings or :class:`~p4p.Type`.

    """
    def wrapper(fn):
        fn._call_PV = pvname
        fn._call_Request = request
        fn._reply_Type = rtype
        return fn
    return wrapper


class RPCDispatcherBase(DynamicProvider):
    def __init__(self, queue, target=None, channels=set(), name=None):
        DynamicProvider.__init__(self, name, self)  # we are our own Handler
        self.queue = queue
        self.target = target
        self.channels = set(channels)
        self.name = name
        self.__pv = SharedPV(
            handler=self,  # no per-channel state, and only RPC used, so only need on PV
            initial=Value(Type([])),  # we don't support get/put/monitor, so use empty struct
        )
        M = self.methods = {}
        for name, mem in inspect.getmembers(target):
            if not hasattr(mem, '_reply_Type'):
                continue
            M[name] = mem

    def getMethodNameArgs(self, request):
        raise NotImplementedError("Sub-class must implement getMethodName")
        # sub-class needs to extract method name from request
        # return 'name', {'var':'val'}

    def testChannel(self, name):
        _log.debug("Test RPC channel %s = %s", name, name in self.channels)
        return name in self.channels

    def makeChannel(self, name, src):
        if self.testChannel(name):
            _log.debug("Open RPC channel %s", name)
            return self.__pv  # no per-channel tracking needed
        else:
            _log.warn("Ignore RPC channel %s", name)

    def rpc(self, pv, op):
        _log.debug("RPC call %s", op)
        try:
            self.queue.push(partial(self._handle, op))
        except Full:
            _log.warn("RPC call queue overflow")
            op.done(error="Too many concurrent RPC calls")

    def _handle(self, op):
        try:
            request = op.value()
            name, args = self.getMethodNameArgs(request)
            fn = self.methods[name]
            rtype = fn._reply_Type

            R = fn(**args)

            if not isinstance(R, Value):
                try:
                    R = Value(rtype, R)
                except:
                    _log.exception("Error encoding %s as %s", R, rtype)
                    op.done(error="Error encoding reply")
                    return
            _log.debug("RPC reply %s -> %r", request, R)
            op.done(R)

        except RemoteError as e:
            _log.debug("RPC reply %s -> error: %s", request, e)
            op.done(error=str(e))

        except:
            _log.exception("Error handling RPC %s", request)
            op.done(error="Error handling RPC")


class NTURIDispatcher(RPCDispatcherBase):

    """RPC dispatcher using NTURI (a al. eget)

    Method names are prefixed with a fixed string.

    >>> queue = WorkQueue()
    >>> class Summer(object):
        @rpc([('result', 'i')])
        def add(self, a=None, b=None):
            return {'result': int(a)+int(b)}
    >>> installProvider("arbitrary", NTURIDispatcher(queue, target=Summer(), prefix="pv:prefix:"))

    Making a call with the CLI 'eget' utility::

      $ eget -s pv:prefix:add -a a=1 -a b=2
      ....
      int result 3

    :param queue WorkQueue: A WorkQueue to which RPC calls will be added
    :param prefix str: PV name prefix used by RPC methods
    :param target: The object which has the RPC calls
    """

    def __init__(self, queue, prefix=None, **kws):
        RPCDispatcherBase.__init__(self, queue, **kws)
        self.prefix = prefix
        self.methods = dict([(prefix + meth, fn) for meth, fn in self.methods.items()])
        self.channels = set(self.methods.keys())
        _log.debug('NTURI methods: %s', ', '.join(self.channels))

    def getMethodNameArgs(self, request):
        # {'schema':'pva', 'path':'pvname', 'query':{'var':'val', ...}}
        return request.path, dict(request.query.items())

# legecy for MASAR only
# do not use in new code


class MASARDispatcher(RPCDispatcherBase):

    def __init__(self, queue, **kws):
        RPCDispatcherBase.__init__(self, queue, **kws)
        _log.debug("MASAR pv %s methods %s", self.channels, self.methods)

    def getMethodNameArgs(self, request):
        # all through a single PV, method name in request
        # {'function':'rpcname', 'name':['name', ...], 'value':['val', ...]}
        return request.function, dict(zip(request.get('name', []), request.get('value', [])))


def quickRPCServer(provider, prefix, target,
                   maxsize=20,
                   workers=1,
                   useenv=True, conf=None, isolate=False):
    """Run an RPC server in the current thread

    Calls are handled sequentially, and always in the current thread, if workers=1 (the default).
    If workers>1 then calls are handled concurrently by a pool of worker threads.
    Requires NTURI style argument encoding.

    :param str provider: A provider name.  Must be unique in this process.
    :param str prefix: PV name prefix.  Along with method names, must be globally unique.
    :param target: The object which is exporting methods.  (use the :func:`rpc` decorator)
    :param int maxsize: Number of pending RPC calls to be queued.
    :param int workers: Number of worker threads (default 1)
    :param useenv: Passed to :class:`~p4p.server.Server`
    :param conf: Passed to :class:`~p4p.server.Server`
    :param isolate: Passed to :class:`~p4p.server.Server`
    """
    from p4p.server import Server
    import time
    queue = ThreadedWorkQueue(maxsize=maxsize, workers=workers)
    provider = NTURIDispatcher(queue, target=target, prefix=prefix, name=provider)
    threads = []
    server = Server(providers=[provider], useenv=useenv, conf=conf, isolate=isolate)
    with server, queue:
        while True:
            time.sleep(10.0)


class RPCProxyBase(object):

    """Base class for automatically generated proxy classes
    """
    context = None
    "The Context provided on construction"
    format = None
    "The tuple/dict used to format ('%' operator) PV name strings."
    timeout = 3.0
    "Timeout of RPC calls in seconds"
    authority = ''
    "Authority string sent with NTURI requests"
    throw = True
    "Whether call errors raise an exception, or return it"
    scheme = None  # set to override automatic


def _wrapMethod(name, meth):
    pv, req = meth._call_PV, meth._call_Request
    if sys.version_info >= (3, 0):
        S = inspect.getfullargspec(meth)
        keywords = S.varkw
    else:
        S = inspect.getargspec(meth)
        keywords = S.keywords
    defaults = S.defaults or [] # for getfullargspec().defaults can be None

    if S.varargs is not None or keywords is not None:
        raise TypeError("vararg not supported for proxy method %s" % name)

    if len(S.args) != len(defaults):
        raise TypeError("proxy method %s must specify types for all arguments" % name)

    try:
        NT = NTURI(zip(S.args, defaults))
    except Exception as e:
        raise TypeError("%s : failed to build method from %s, %s" % (e, S.args, defaults))

    @wraps(meth)
    def mcall(self, *args, **kws):
        pvname = pv % self.format
        try:
            uri = NT.wrap(pvname, args, kws, scheme=self.scheme or self.context.name, authority=self.authority)
        except Exception as e:
            raise ValueError("Unable to wrap %s %s as %s (%s)" % (args, kws, NT, e))
        return self.context.rpc(pvname, uri, request=req, timeout=self.timeout, throw=self.throw)

    return mcall


def rpcproxy(spec):
    """Decorator to enable this class to proxy RPC client calls

    The decorated class constructor takes two additional arguments,
    `context=` is required to be a :class:`~p4p.client.thread.Context`.
    `format`= can be a string, tuple, or dictionary and is applied
    to PV name strings given to :py:func:`rpcall`.
    Other arguments are passed to the user class constructor. ::

       @rpcproxy
       class MyProxy(object):
           @rpccall("%s:add")
           def add(lhs='d', rhs='d'):
               pass

       ctxt = Context('pva')
       proxy = MyProxy(context=ctxt, format="tst:")  # evaluates "%s:add"%"tst:"

    The decorated class will be a sub-class of the provided class and :class:`RPCProxyBase`.
    """
    # inject our ctor first so we don't have to worry about super() non-sense.

    def _proxyinit(self, context=None, format={}, **kws):
        assert context is not None, context
        self.context = context
        self.format = format
        spec.__init__(self, **kws)
    obj = {'__init__': _proxyinit}

    for K, V in inspect.getmembers(spec, lambda M: hasattr(M, '_call_PV')):
        obj[K] = _wrapMethod(K, V)

    return type(spec.__name__, (RPCProxyBase, spec), obj)



#cython: language_level=2

from libcpp cimport bool
from libcpp.string cimport string
from libcpp.memory cimport shared_ptr
from libcpp.set cimport set as setxx

from .data cimport Value

cdef extern from "<pvxs/source.h>" namespace "pvxs::server" nogil:
    cdef cppclass Source:
        pass

    enum op_t "pvxs::server::OpBase::op_t":
        None "pvxs::server::OpBase::None",
        Info "pvxs::server::OpBase::Info",
        Get "pvxs::server::OpBase::Get",
        Put "pvxs::server::OpBase::Put",
        RPC "pvxs::server::OpBase::RPC",

    cdef cppclass ClientCredentials:
        string peer
        string iface
        string method
        string account
        Value raw
        setxx[string] roles() except+

    cdef cppclass OpBase:
        const string& peerName() const
        const string& name() const
        const shared_ptr[const ClientCredentials]& credentials() const

    cdef cppclass ExecOp(OpBase):
        void reply() except+
        void reply(const Value&) except+
        void error(const string&) except+
        Value pvRequest() const

    cdef cppclass ChannelControl(OpBase):
        void close() except+

    cdef cppclass ClientCredentials:
        string peer
        string iface
        string method
        string account
        Value raw
        setxx[string] roles() except+



# I'm imported from test_asyncio.y
# Add new TestCase s to __all__

import logging
_log = logging.getLogger(__name__)

from functools import wraps

from .. import Value
from ..nt import NTScalar
from ..server import Server, StaticProvider
from .utils import RefTestCase

import asyncio

from ..client.asyncio import Context, Disconnected, timesout
from ..server.asyncio import SharedPV

__all__ = (
    'TestGPM',
    'TestTimeout',
    'TestFirstLast',
)

# we should never implicitly use the default loop.
asyncio.get_event_loop().close()

class AsyncMeta(type):
    """Automatically wrap and "async def test*():" methods for dispatch to self.loop
    """
    def __new__(klass, name, bases, classdict):
        for name, mem in classdict.items():
            if name.startswith('test') and asyncio.iscoroutinefunction(mem):
                @wraps(mem)
                def wrapper(self, mem=mem):
                    self.loop.run_until_complete(asyncio.wait_for(mem(self), self.timeout))
                classdict[name] = wrapper

        return type.__new__(klass, name, bases, classdict)

class AsyncTest(RefTestCase, metaclass=AsyncMeta):
    timeout = 1.0

    def setUp(self):
        super(AsyncTest, self).setUp()
        self.loop = asyncio.new_event_loop()
        self.loop.set_debug(True)
        self.loop.run_until_complete(asyncio.wait_for(self.asyncSetUp(), self.timeout))

    def tearDown(self):
        self.loop.run_until_complete(asyncio.wait_for(self.asyncTearDown(), self.timeout))
        self.loop.close()
        super(AsyncTest, self).tearDown()

    async def asyncSetUp(self):
        pass

    async def asyncTearDown(self):
        pass

class TestGPM(AsyncTest):
    timeout = 3  # overall timeout for each test method

    class Handler:
        async def put(self, pv, op):
            _log.debug("putting %s <- %s", op.name(), op.value())
            await asyncio.sleep(0)  # prove that we can
            pv.post(op.value() * 2)
            op.done()

    async def asyncSetUp(self):
        await super(TestGPM, self).asyncSetUp()

        self.pv = SharedPV(nt=NTScalar('i'), initial=0, handler=self.Handler())
        self.pv2 = SharedPV(handler=self.Handler(), nt=NTScalar('d'), initial=42.0)
        self.provider = StaticProvider("serverend")
        self.provider.add('foo', self.pv)
        self.provider.add('bar', self.pv2)

    async def asyncTearDown(self):
        del self.pv
        del self.pv2
        del self.provider
        await super(TestGPM, self).asyncTearDown()

    async def test_getput(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:
                self.assertEqual(0, (await C.get('foo')))

                await C.put('foo', 5)

                self.assertEqual(5 * 2, (await C.get('foo')))

    async def test_monitor(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:

                Q = asyncio.Queue()

                sub = C.monitor('foo', Q.put, notify_disconnect=True)
                try:
                    self.assertIsInstance((await Q.get()), Disconnected)

                    self.assertEqual(0, (await Q.get()))

                    await C.put('foo', 2)

                    self.assertEqual(2 * 2, (await Q.get()))

                    self.pv.close()

                    self.assertIsInstance((await Q.get()), Disconnected)

                    self.pv.open(3)

                    self.assertEqual(3, (await Q.get()))

                finally:
                    sub.close()
                    await sub.wait_closed()

    async def test_put_noconvert(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:
                with self.assertRaisesRegex(ValueError, 'not_an_integer'):
                    await C.put('foo', 'not_an_integer')

    if hasattr(asyncio, 'coroutine'):
        @asyncio.coroutine
        def test_gen_coro(self):
            # demonstrate interoperability w/ code using generated based coroutines
            with Server(providers=[self.provider], isolate=True) as S:
                with Context('pva', conf=S.conf(), useenv=False) as C:
                    self.assertEqual(0, (yield from C.get('foo')))

                    yield from C.put('foo', 5)

                    self.assertEqual(5 * 2, (yield from C.get('foo')))


class TestTimeout(AsyncTest):

    async def test_timeout(self):
        done = None

        @timesout()
        async def action():
            nonlocal done
            done = False
            await asyncio.sleep(5)
            done = True

        with self.assertRaises(asyncio.TimeoutError):
            await action(timeout=0.01)

class TestFirstLast(AsyncTest):
    maxDiff = 2000
    timeout = 5.0
    mode = 'Mask'

    class Handler:
        def __init__(self):
            self.evtC = asyncio.Event()
            self.evtD = asyncio.Event()
            self.conn = None
        def onFirstConnect(self, pv):
            _log.debug("onFirstConnect")
            self.conn = True
            self.evtC.set()
        def onLastDisconnect(self, pv):
            _log.debug("onLastDisconnect")
            self.conn = False
            self.evtD.set()

    async def asyncSetUp(self):
        await super(TestFirstLast, self).asyncSetUp()

        self.H = self.Handler()
        self.pv = SharedPV(handler=self.H,
                           nt=NTScalar('d'),
                           options={'mapperMode':self.mode})
        self.sprov = StaticProvider("serverend")
        self.sprov.add('foo', self.pv)

        self.server = Server(providers=[self.sprov], isolate=True)

    async def asyncTearDown(self):
        self.server.stop()
        #_defaultWorkQueue.stop()
        del self.server
        del self.sprov
        del self.pv
        del self.H
        await super(TestFirstLast, self).asyncTearDown()

    async def testClientDisconn(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = asyncio.Queue(maxsize=4)
            sub = ctxt.monitor('foo', Q.put, notify_disconnect=True)
            try:

                _log.debug('Wait Disconnected')
                E = await Q.get() # Disconnected
                self.assertIsInstance(E, Disconnected)
                _log.debug('Wait initial')
                E = await Q.get() # initial update
                self.assertIsInstance(E, Value)

                _log.debug('Wait onFirstConnect')
                await self.H.evtC.wait() # onFirstConnect()
                self.assertTrue(self.H.conn)

            except:
                _log.exception('oops')
                raise
            finally:
                _log.debug('sub close()')
                sub.close()
                await sub.wait_closed()

        _log.debug('Wait onLastDisconnect')
        await self.H.evtD.wait() # onLastDisconnect()
        _log.debug('SHUTDOWN')
        self.assertFalse(self.H.conn)

    async def testServerShutdown(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = asyncio.Queue(maxsize=4)
            sub = ctxt.monitor('foo', Q.put, notify_disconnect=True)
            try:

                await Q.get() # initial update

                _log.debug('TEST')
                await self.H.evtC.wait() # onFirstConnect()
                self.assertIs(self.H.conn, True)

                self.server.stop()

                await self.H.evtD.wait() # onLastDisconnect()
                _log.debug('SHUTDOWN')
                self.assertIs(self.H.conn, False)

            finally:
                sub.close()
                await sub.wait_closed()

    async def testPVClose(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = asyncio.Queue(maxsize=4)
            sub = ctxt.monitor('foo', Q.put, notify_disconnect=True)
            try:

                await Q.get() # initial update

                _log.debug('TEST')
                await self.H.evtC.wait() # onFirstConnect()
                self.assertTrue(self.H.conn)

                await self.pv.close(destroy=True, sync=True) # onLastDisconnect()

                _log.debug('CLOSE')
                self.assertFalse(self.H.conn)

            finally:
                sub.close()
                await sub.wait_closed()



"""
See https://github.com/epics-extensions/ca-gateway/blob/master/docs/GATEWAY.pvlist
"""

import socket
import logging
import re
from collections import defaultdict, OrderedDict

_log = logging.getLogger(__name__)

def _sub_add(expr, ngroups, adjust=0):
    '''Adjust RE substitution offsets
    '''
    def check(M):
        grp = int(M.group(1))
        if grp<1 or grp>ngroups:
            raise ValueError('Out of range substitution %d.  Must be in [1, %d]'%(grp,ngroups))
        return r'\%d'%(grp+adjust)
    return re.sub(r'(?<!\\)\\([0-9]+)', check, expr)

def _re_join(exprs, capture=''):
    A = '|'.join(['(%s%s)'%(capture, E) for E in exprs])
    return re.compile('^(?:%s)$'%A)

class PVList(object):
    '''Parse and prepare pvlist text.

    >>> with open(fname, 'r') as F:
            pvl = PVList(F.read())
    '''
    def __init__(self, pvl):
        allowfirst = False

        # {'host':[RE, ...]}
        deny_from = defaultdict(set)
        deny_all = set()
        allow = OrderedDict() # {RE:(sub|None, asg, asl)}
        # number of match groups encountered so far.
        # must match between allow key (pattern) and substitution
        ngroups = 1 # one indexed
        self._allow_groups = []

        lines = (pvl or '.* ALLOW').splitlines()
        # ALLOW entries are given in order of increasing precedence.
        # The last match in the file is used.
        lines.reverse()

        lineno = 0
        for line in lines:
            lineno += 1
            try:
                line = line.strip()
                if len(line)==0 or line[:1]=='#':
                    continue

                M = re.match(r'^\s*EVALUATION\s+ORDER\s+([A-Z]+),\s+([A-Z]+)\s*$', line)
                if M:
                    if M.groups()==('DENY', 'ALLOW'):
                        allowfirst = True # allow rules take precedence
                        _log.warn('Ignoring "EVALUATION ORDER DENY, ALLOW".  Only ALLOW, DENY is implemented.')
                    elif M.groups()==('ALLOW', 'DENY'):
                        allowfirst = False # deny rules take precedence (default)
                    else:
                        raise RuntimeError("Invalid order: %s"%(M.groups(),))
                    continue

                parts = [part.strip() for part in line.split(None)]
                pattern, cmd, parts = parts[0], parts[1].upper(), parts[2:]

                # test compile
                C = re.compile(pattern)

                if pattern in allow:
                    continue # ignore duplicate pattern

                if cmd=='DENY':
                    if len(parts) and parts[0].upper()=='FROM':
                        parts = parts[1:]

                    if parts:
                        for host in parts:
                            host = self._gethostbyname(host)
                            deny_from[host].add(pattern)

                    else:
                        deny_all.add(pattern)

                elif cmd=='ALIAS':
                    self._allow_groups.append(ngroups)
                    ngroups += 1 # _re_join adds one capture group

                    alias = _sub_add(parts[0], ngroups=C.groups, adjust=ngroups-1)
                    asg = parts[1] if len(parts)>1 else 'DEFAULT'
                    asl = int(parts[2] if len(parts)>2 else '0')

                    allow[pattern] = (alias, asg, asl)
                    ngroups += C.groups

                elif cmd=='ALLOW':
                    self._allow_groups.append(ngroups)
                    ngroups += 1 # _re_join adds one capture group

                    asg = parts[0] if len(parts)>0 else 'DEFAULT'
                    asl = int(parts[1] if len(parts)>1 else '0')

                    allow[pattern] = (None, asg, asl)

                else:
                    raise RuntimeError("Unknown command: %s"%cmd)

            except Exception as e:
                raise e.__class__("Error on line %s: %s"%(lineno, e))

        deny_all = list(deny_all)

        # RE's for each host specific list also include the general list.
        # So only need to run one deny RE for request
        self._deny_from = {addr:_re_join(list(exprs)+deny_all, '?:') for addr, exprs in deny_from.items()}
        self._deny_all = _re_join(deny_all, '?:')

        allow_pat, self._allow_actions = list(allow.keys()), list(allow.values())

        self._allow_pat = _re_join(allow_pat)

        assert self._allow_pat.groups+1==ngroups, (self._allow_pat.groups, ngroups)

    @staticmethod
    def _gethostbyname(host):
        return socket.gethostbyname(host)

    def compute(self, pv, addr):
        '''Lookup PV name and client IP address.

        Returns a triple of None/rewritten PV name, security group name (ASG), and security level number (ASL).
        '''
        pv = pv.decode('UTF-8')
        P = self._deny_from.get(addr, self._deny_all)

        if not P.match(pv):
            M = self._allow_pat.match(pv)
            if M:
                for idx, val in enumerate(M.group(*self._allow_groups)):
                    if val is not None:
                        alias, asg, asl = self._allow_actions[idx]
                        if alias is not None:
                            pv = M.expand(alias)

                        return pv, asg, asl

        return None, None, None



import logging
from math import ceil

from qtpy.QtCore import QObject, QCoreApplication, Signal, QEvent

from . import raw
from .raw import Disconnected, RemoteError, Cancelled, Finished
from ..wrapper import Value, Type
from .._p4p import ClientProvider
from .._p4p import (logLevelAll, logLevelTrace, logLevelDebug,
                    logLevelInfo, logLevelWarn, logLevelError,
                    logLevelFatal, logLevelOff)

from .thread import TimeoutError

__all__ = (
    'Context',
    'Value',
    'Type',
    'RemoteError',
    'TimeoutError',
)

_log = logging.getLogger(__name__)

# some pyqt callbacks are delicate, and will SIGSEGV is a python exception is allowed to propagate
def exceptionGuard(fn):
    def wrapper(*args, **kws):
        try:
            fn(*args, **kws)
        except:
            _log.exception('oops')
    return wrapper

class CBEvent(QEvent):
    # allocate an event ID#
    EVENT_TYPE = QEvent.Type(QEvent.registerEventType())

    def __init__(self, result):
        QEvent.__init__(self, self.EVENT_TYPE)
        self.result = result

class Operation(QObject):
    _op = None
    _result = None
    # receives a Value or an Exception
    result = Signal(object)

    def __init__(self, parent, timeout):
        QObject.__init__(self, parent)
        self._active = self.startTimer(ceil(timeout*1000))

    def close(self):
        self._op.close()

    @exceptionGuard
    def timerEvent(self, evt):
        if self._result is None:
            self._result = TimeoutError()
            self.result.emit(self._result)

    def _resultcb(self, value):
        # called on PVA worker thread
        QCoreApplication.postEvent(self, CBEvent(value))

    @exceptionGuard
    def customEvent(self, evt):
        if self._active is not None:
            self.killTimer(self._active)
            self._active = None

        if self._result is None:
            self._result = evt.result
            self.result.emit(self._result)

class MCache(QObject):
    _op = None
    # receives a Value or an Exception
    update = Signal(object)

    def __init__(self, name, parent):
        QObject.__init__(self, parent)

        self.name = name
        self._active = None
        self._holdoff = 10*1000 # acts as low limit on high limit
        self._last = Disconnected()

    def _add(self, slot, limitHz=10.0):
        holdoff = int(max(0.1, 1.0/limitHz)*1000)

        # Rate limiting for multiple consumers is hard.
        # We throttle to the highest consumer rate (shortest holdoff).
        if self._holdoff is None or self._holdoff > holdoff:
            self._holdoff = holdoff
            if self._active is not None:
                # restart timer
                self.killTimer(self._active)
                self._active = self.startTimer(ceil(self._holdoff))

        # TODO: re-adjust on slot disconnect?

        # schedule to receive initial update later (avoids recursion)
        QCoreApplication.postEvent(self, CBEvent(slot))

    def _event(self):
        _log.debug('event1 %s', self.name)
        # called on PVA worker thread

        QCoreApplication.postEvent(self, CBEvent(None))

    @exceptionGuard
    def customEvent(self, evt):
        E = evt.result
        _log.debug('event2 %s %s', self.name, E)
        # E will be one of:
        #   None - FIFO not empty (call pop())
        #   RemoteError
        #   Disconnected
        #   some method, adding new subscriber

        if E is None:
            if self._active is None:
                self._active = self.startTimer(ceil(self._holdoff))
                _log.debug('Start timer with %s', self._holdoff)
            return

        else:
            E(self._last)
            self.update.connect(E)

    @exceptionGuard
    def timerEvent(self, evt):
        V = self._op.pop()
        _log.debug('tick %s %s', self.name, V)

        if V is not None:
            self._last = V
            self.update.emit(V)

        elif self._active is not None:
            self.killTimer(self._active)
            self._active = None

class Context(raw.Context):
    """Context(provider, conf=None, useenv=True)
    
    PyQt aware Context.
    Methods in the class give notification of completion/update through Qt signals.

    :param str provider: A Provider name.  Try "pva" or run :py:meth:`Context.providers` for a complete list.
    :param dict conf: Configuration to pass to provider.  Depends on provider selected.
    :param bool useenv: Allow the provider to use configuration from the process environment.
    :param dict nt: Controls :ref:`unwrap`.  None uses defaults.  Set False to disable
    :param dict unwrap: Legacy :ref:`unwrap`.
    :param parent QObject: Parent for QObjects created through this Context.
    """
    def __init__(self, provider, parent=None, **kws):
        super(Context, self).__init__(provider, **kws)
        self._parent = QObject(parent)

        self._mcache = {}
        self._puts = {}


    def disconnect(self, name=None):
        if name is None:
            self._mcache = {}
            self._puts = {}
        else:
            self._mcache.pop(name)
            self._puts.pop(name)
        super(Context, self).disconnect(name)

    # get() omitted (why would a gui want to do this?)

    def put(self, name, value, slot=None, request=None, timeout=5.0,
            process=None, wait=None, get=True):
        """Write a new value to a single PV.

        Returns an Operation instance which will emit either a success and error signal.
        If the slot argument is provided, this will be connected in a race free way.

        The slot function will receive a python object which is either None (Success) or an Exception.

        Note that the returned Operation will also be stored internally by the Context.
        So the caller is not required to store it as well.
        This internal storage will only keep the most recent put() Operation for each PV name.
        A previous incomplete put() will be cancelled if/when put() is called again.

        :param name: A single name string or list of name strings
        :param values: A single value, a list of values, a dict, a `Value`.  May be modified by the constructor nt= argument.
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param slot: A callable object, such as a bound method, which can be passed to QObject.connect()
        :param float timeout: Operation timeout in seconds
        :param str process: Control remote processing.  May be 'true', 'false', 'passive', or None.
        :param bool wait: Wait for all server processing to complete.
        :param bool get: Whether to do a Get before the Put.  If True then the value passed to the builder callable
                         will be initialized with recent PV values.  eg. use this with NTEnum to find the enumeration list.
        """
        if request and (process or wait is not None):
            raise ValueError("request= is mutually exclusive to process= or wait=")
        elif process or wait is not None:
            request = 'field()record[block=%s,process=%s]' % ('true' if wait else 'false', process or 'passive')

        prev = self._puts.get(name)
        if prev is not None:
            # issuing new Put implicitly cancels any pending/queued Put
            prev.close()

        self._puts[name] = op = Operation(self._parent, timeout)
        if slot is not None:
            op.result.connect(slot)

        op._op = super(Context, self).put(name, op._resultcb, builder=value, request=request, get=get)

        return op

    def rpc(self, name, value, slot, request=None, timeout=5.0, throw=True):
        """Perform a Remote Procedure Call (RPC) operation

        Returns an Operation instance which will emit either a success and error signal to the provided slot.
        This Operation instance must be stored by the caller or it will be implicitly cancelled.

        The slot function will receive a python object which is either a Value or an Exception.

        :param str name: PV name string
        :param Value value: Arguments.  Must be Value instance
        :param slot: A callable object, such as a bound method, which can be passed to QObject.connect()
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param float timeout: Operation timeout in seconds
        :param bool throw: When true, operation error throws an exception.
                     If False then the Exception is returned instead of the Value

        :returns: An Operation
        """

        op = Operation(self._parent, timeout)
        op.result.connect(slot)
        op._op = super(Context, self).rpc(name, op._resultcb, value, request=request)
        return op

    def monitor(self, name, slot, request=None, limitHz=10.0):
        """Request subscription to named PV

        Request notification to the provided slot when a PV is updated.
        Subscriptions are managed by an internal cache,
        so than multiple calls to monitor() with the same PV name may be satisfied through a single subscription.

        limitHz, which must be provided, puts an upper limit on the rate at which the update signal will be emitted.
        Some update will be dropped in the PV updates more frequently.
        Reduction is done by discarding the second to last update.
        eg. It is guaranteed that the last update (present value) in the burst will be delivered.

        :param str name: PV name string
        :param callable cb: Processing callback
        :param slot: A callable object, such as a bound method, which can be passed to QObject.connect()
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param limitHz float: Maximum rate at which signals will be emitted.  In signals per second.
        :returns: a :py:class:`MCache` instance
        """
        _log.debug('Subscribe to %s with %s', name, request)
        if request is not None:
            raise NotImplementedError("monitor() with request= not yet implemented")
        #if isinstance(request, (str, bytes)):
        #    request = ClientProvider.makeRequest(request)
        #if isinstance(request, Value):
        #    request = serialize(request)

        key = (name, request) # (str, bytes|None)

        try:
            op = self._mcache[key]
        except KeyError:
            self._mcache[key] = op = MCache(name, self._parent)

            op._op = super(Context, self).monitor(name, op._event, request)

        op._add(slot, limitHz=limitHz)

        return op




import weakref
import gc
import unittest
from collections import OrderedDict

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_aequal

from ..wrapper import Type, Value
from .. import pvdVersion
from .utils import RefTestCase


class TestWrapper(RefTestCase):
    def testCall(self):
        T = Type([('ival', 'I')])
        V = T({'ival':4})
        self.assertEqual(V.ival, 4)

class TestRawValue(RefTestCase):

    def testToString(self):
        V = Value(Type([
            ('ival', 'i'),
            ('dval', 'd'),
            ('sval', 's'),
        ]), {
            'ival': 42,
            'dval': 4.2,
            'sval': 'hello',
        })

        # attempt to normalize to avoid comparing whitespace
        def proc(s):
            return list([l.split() for l in s.splitlines()])

        self.assertListEqual(proc(str(V)),
                         proc('''struct {
    int32_t ival = 42
    double dval = 4.2
    string sval = "hello"
}
'''))

    def testScalar(self):
        V = Value(Type([
            ('ival', 'i'),
            ('dval', 'd'),
            ('sval', 's'),
        ]), {
            'ival': 42,
            'dval': 4.2,
            'sval': 'hello',
        })

        self.assertListEqual(V.tolist(), [
            ('ival', 42),
            ('dval', 4.2),
            ('sval', u'hello'),
        ])

        self.assertListEqual(V.items(), [
            ('ival', 42),
            ('dval', 4.2),
            ('sval', u'hello'),
        ])

        self.assertEqual(V.ival, 42)
        self.assertEqual(V.dval, 4.2)
        self.assertEqual(V.sval, u'hello')

        self.assertEqual(V['ival'], 42)
        self.assertEqual(V['dval'], 4.2)
        self.assertEqual(V['sval'], u'hello')

        V.ival = 43
        self.assertEqual(V.ival, 43)
        self.assertEqual(V['ival'], 43)

        V.dval = 4.3
        self.assertEqual(V.dval, 4.3)
        self.assertEqual(V['dval'], 4.3)

        V.sval = u'world'
        self.assertEqual(V.sval, u'world')
        self.assertEqual(V['sval'], u'world')

    def testIntegerRange(self):
        # test limits of intger ranges
        V = Value(Type([
            ('i32', 'i'),
            ('u32', 'I'),
            ('i64', 'l'),
            ('u64', 'L'),
        ]), {
            'i32': 0x7fffffff,
            'u32': 0xffffffff,
            'i64': 0x7fffffffffffffff,
            'u64': 0xffffffffffffffff,
        })

        self.assertEqual(V.i32, 0x7fffffff)
        self.assertEqual(V.u32, 0xffffffff)
        self.assertEqual(V.i64, 0x7fffffffffffffff)
        self.assertEqual(V.u64, 0xffffffffffffffff)

        V.i32 = -0x80000000
        V.i64 = -0x8000000000000000

        self.assertEqual(V.i32, -0x80000000)
        self.assertEqual(V.i64, -0x8000000000000000)

        # test setting out of range
        # TODO: not great...

        V.i64 = 0x8000000000000000
        self.assertEqual(V.i64, -0x8000000000000000)

        with self.assertRaises(OverflowError):
            V.i64 = 0x10000000000000000
        with self.assertRaises(OverflowError):
            V.i64 = -0x8000000000000001
        with self.assertRaises(OverflowError):
            V.u64 = 0x10000000000000000

        V.i32 = 0x80000000
        self.assertEqual(V.i32, -0x80000000)

        V.i32 = 0x100000000
        self.assertEqual(V.i32, 0)

    def testFieldAccess(self):
        V = Value(Type([
            ('ival', 'i'),
            ('dval', 'd'),
            ('sval', 's'),
        ]), {
            'ival': 42,
            'dval': 4.2,
            'sval': 'hello',
        })

        self.assertEqual(100, V.get('foo', 100))
        self.assertIsNone(V.get('foo'))

        self.assertRaises(KeyError, V.__getitem__, 'foo')
        self.assertRaises(AttributeError, getattr, V, 'foo')

        self.assertRaises(KeyError, V.__setitem__, 'foo', 5)
        self.assertRaises(AttributeError, setattr, V, 'foo', 5)

    def testReserved(self):
        L = [
            ("name", "s"),
            ("field", "I"),
            ("type", "s"),
        ]
        T = Type(L)
        V = Value(T)

        # item access works as always
        V['type'] = 'hello'
        self.assertEqual(V['type'], 'hello')

        # 'type' field hidden by method
        T2 = V.type() # ensure Value.type() isn't hidden

        # can't prevent overwriting the method...
        V.type = 4

    def testBadField(self):
        T = Type([
            ('ival', 'i'),
            ('dval', 'd'),
            ('sval', 's'),
        ])
        self.assertRaises(KeyError, Value, T, {'invalid': 42})

    def testArray(self):
        V = Value(Type([
            ('ival', 'ai'),
            ('dval', 'ad'),
            ('sval', 'as'),
        ]), {
            'ival': [1, 2, 3],
            'dval': np.asarray([1.1, 2.2], dtype=np.float64),
            'sval': ['a', u'b'],
        })

        assert_aequal(V.ival, np.asarray([1, 2, 3]))
        assert_aequal(V.dval, np.asarray([1.1, 2.2], dtype=np.float64))
        self.assertListEqual(V.sval, [u'a', u'b'])

        # i4 <- u8 violates "safe" casting rules
        V.ival = np.asarray([4, 5], dtype='u8')
        assert_aequal(V.ival, np.asarray([4, 5]))

    def testSubStruct(self):
        V = Value(Type([
            ('ival', 'i'),
            ('str', ('S', 'foo', [
                ('a', 'i'),
                ('b', 'i'),
            ])),
        ]), {
            'ival': 42,
            'str': {
                'a': 1,
                'b': 2,
            },
        })

        self.assertListEqual(V.tolist(), [
            ('ival', 42),
            ('str', [
                ('a', 1),
                ('b', 2),
            ]),
        ])

        self.assertDictEqual(V.todict(), {
            'ival': 42,
            'str': {
                'a': 1,
                'b': 2,
            },
        })

        D = V.todict(None, OrderedDict)
        self.assertIsInstance(D, OrderedDict)
        self.assertDictEqual(D, {
            'ival': 42,
            'str': {
                'a': 1,
                'b': 2,
            },
        })

        self.assertListEqual(V.tolist('str'), [
            ('a', 1),
            ('b', 2),
        ])

        self.assertEqual(V.str.a, 1)
        self.assertEqual(V.str['a'], 1)
        self.assertEqual(V['str'].a, 1)
        self.assertEqual(V['str']['a'], 1)
        self.assertEqual(V['str.a'], 1)

        self.assertEqual(V.type().aspy(),
                         ('S', 'structure', [
                          ('ival', 'i'),
                          ('str', ('S', 'foo', [
                                   ('a', 'i'),
                                   ('b', 'i'),
                                   ])),
                          ]),
                         )

        self.assertEqual(V.type('str').aspy(),
                         ('S', 'foo', [
                          ('a', 'i'),
                          ('b', 'i'),
                          ]),
                         )

        self.assertRaises(KeyError, V.type, 'invalid')

        self.assertListEqual(V.keys(), ['ival', 'str'])
        self.assertTrue('ival' in V)
        self.assertTrue('str' in V)
        self.assertFalse('missing' in V)

    def testVariantUnion(self):
        V = Value(Type([
            ('x', 'v'),
        ]))

        self.assertIsNone(V.x)

        V.x = True
        self.assertIs(V.x, True)

        V.x = 5
        self.assertEqual(V.x, 5)

        V.x = 1.1
        self.assertEqual(V.x, 1.1)

        V.x = 'test'
        self.assertEqual(V.x, u'test')

        V.x = u'test'
        self.assertEqual(V.x, u'test')

        V.x = np.asarray([1, 2])
        assert_aequal(V.x, np.asarray([1, 2]))

        V.x = np.asarray([1, 2], dtype=np.float64)
        assert_aequal(V.x, np.asarray([1, 2], dtype=np.float64))

        # clearing unions is broken prior to 7.0.0
        if pvdVersion() >= (7, 0, 0, 0):
            V.x = None
            self.assertIsNone(V.x)

    def testVariantUnionExplicit(self):
        V = Value(Type([
            ('x', 'v'),
        ]))

        V.x = ('B', 0xffff)
        self.assertEqual(V.x, 0xff)

        V.x = ('d', 42)
        self.assertEqual(V.x, 42.0)
        self.assertEqual(type(V.x), float)

        V.x = ('ad', np.asarray([1, 2]))
        assert_aequal(V.x, np.asarray([1, 2], dtype=np.float64))

        V.x = ('as', ['one', 'two'])
        self.assertEqual(V.x, ['one', 'two'])

    def testVariantUnionRoundTrip(self):
        "Ensure type is preserved"
        T = Type([
            ('x', 'v'),
        ])
        V = T()

        for X in (None,
                  True,
                  42,
                  4.2,
                  u"hello",
                  [u"hello", u"world"],
                  ):
            V.x = X
            self.assertEqual(X, V.x)
            self.assertIs(type(X), type(V.x))
            V2 = T()
            V2[None] = V
            self.assertEqual(X, V2.x)
            self.assertIs(type(X), type(V2.x))

        for X in (np.asarray([True, False]),
                  np.asarray([-0x80, 0x7f], dtype='i1'),
                  np.asarray([-0x8000, 0x7fff], dtype='i2'),
                  np.asarray([-0x80000000, 0x7fffffff], dtype='i4'),
                  np.asarray([-0x8000000000000000, 0x7fffffffffffffff], dtype='i8'),
                  np.asarray([0xff], dtype='u1'),
                  np.asarray([0xffff], dtype='u2'),
                  np.asarray([0xffffffff], dtype='u4'),
                  np.asarray([0xffffffffffffffff], dtype='u8'),
                  # float test should not be sensitive to roundoff
                  np.asarray([1.23456789], dtype='f4'),
                  np.asarray([1.23456789e100], dtype='f8'),
                  ):
            try:
                V.x = X
                assert_aequal(V.x, X)
                self.assertEqual(V.x.dtype, X.dtype)
                V2 = T()
                V2[None] = V
                assert_aequal(V2.x, X)
                self.assertEqual(V2.x.dtype, X.dtype)
            except RuntimeError as e:
                raise RuntimeError("With %s : %s"%(X, e))

    def testDisUnion(self):
        V = Value(Type([
            ('x', ('U', 'x', [
                ('a', 'i'),
                ('b', 's'),
            ])),
        ]))

        self.assertIsNone(V.x)

        V.x = ('a', 44)
        self.assertEqual(V.x, 44)

        V.x = ('b', 'test')
        self.assertEqual(V.x, u'test')

        # uses previously selected 'b'
        V.x = 'world'
        self.assertEqual(V.x, u'world')

        # uses previously selected 'b'
        V.x = 128
        self.assertEqual(V.x, u'128')

        # clearing unions is broken prior to 7.0.0
        if pvdVersion() >= (7, 0, 0, 0):
            V.x = None  # another way to clear
            self.assertIsNone(V.x)

    def testUnionArray(self):
        V = Value(Type([
            ('x', 'av'),
            ('y', ('aU', 'foo', [
                ('a', 'i'),
                ('b', 's'),
            ])),
        ]))

        self.assertListEqual(V.x, [])
        self.assertListEqual(V.y, [])

        # variant union assignment
        V.x = [1, 4.2, 'foo']
        # discriminating union automagic selection
        V.y = [2, 5.2, 'bar']

        self.assertListEqual(V.x, [1, 4.2, u'foo'])
        # magic field selection strikes to convert float -> int
        self.assertListEqual(V.y, [2, 5, u'bar'])

        V.y = [('a', 3),
               ('b', 4.2),
               ('b', 'baz')]

        self.assertListEqual(V.y, [3, u'4.2', u'baz'])

        # union array assignment ignores previous selections
        V.y = [2, 5.2, 'bar']
        self.assertListEqual(V.y, [2, 5, u'bar'])

    def testUnionArrayStruct(self):
        S = Value(Type([
            ('x', 'i'),
        ]), {
            'x': 42,
        })

        V = Value(Type([
            ('y', 'av'),
        ]), {
            'y': [S],
        })

        # attribute/item access returns union array w/ struct as list of Value
        X = V.y

        self.assertIsInstance(X, list)
        self.assertIsInstance(X[0], Value)
        self.assertEqual(len(X), 1)

        # returns union array w/ struct as list of list of tuples
        Y = V.tolist()

        self.assertListEqual(Y, [
            ('y', [[
                ('x', 42),
            ]]),
        ])

    def testStructID(self):
        V = Value(Type([('a', 'I')]))
        self.assertEqual(V.getID(), "structure")

        V = Value(Type([('a', 'I')], id="foo"))
        self.assertEqual(V.getID(), "foo")

    def testStructArr(self):
        V = Value(Type([
            ('a', ('aS', None, [
                ('b', 'i'),
            ])),
        ]), {
            'a': [
                {'b': 4},
                [('b', 5)],
            ],
        })

        self.assertEqual(len(V.a), 2)
        self.assertEqual(V.a[0].b, 4)
        self.assertEqual(V.a[1].b, 5)

        V.a = [{'b': 1}]

        self.assertEqual(len(V.a), 1)
        self.assertEqual(V.a[0].b, 1)

    def testBitSet(self):
        A = Value(Type([
            ('x', 'i'),
            ('y', 'i'),
        ]), {
            'y': 42,
        })
        # initially all un-"changed" (at default)

        self.assertSetEqual(A.changedSet(), {'y'})
        self.assertTrue(A.changed())
        self.assertFalse(A.changed('x'))
        self.assertTrue(A.changed('y'))
        self.assertRaises(KeyError, A.changed, 'invalid')

        A.mark('x')

        self.assertSetEqual(A.changedSet(), {'x', 'y'})
        self.assertTrue(A.changed())
        self.assertTrue(A.changed('x'))
        self.assertTrue(A.changed('y'))

        A.mark('x', False)

        self.assertSetEqual(A.changedSet(), {'y'})
        self.assertTrue(A.changed())
        self.assertFalse(A.changed('x'))
        self.assertTrue(A.changed('y'))

        A.unmark()

        self.assertSetEqual(A.changedSet(), set())
        self.assertFalse(A.changed())
        self.assertFalse(A.changed('x'))
        self.assertFalse(A.changed('y'))

    def testBitSetRecurse(self):
        A = Value(Type([
            ('x', 'i'),
            ('y', 'i'),
            ('z', ('S', None, [
                ('a', 'i'),
                ('b', 'i'),
            ])),
        ]), {
        })

        self.assertSetEqual(A.changedSet(), set())
        self.assertFalse(A.changed())
        self.assertFalse(A.changed('x'))
        self.assertFalse(A.changed('y'))
        self.assertFalse(A.changed('z.a'))
        self.assertFalse(A.changed('z.b'))

        A.mark('y')
        A.mark('z.a')
        self.assertSetEqual(A.changedSet(), {'y', 'z.a'})
        self.assertTrue(A.changed())
        self.assertFalse(A.changed('x'))
        self.assertTrue(A.changed('y'))
        self.assertTrue(A.changed('z.a'))
        self.assertFalse(A.changed('z.b'))

        Z = A.z  # A and Z share fields and bitset
        self.assertTrue(Z.changed('a'))
        self.assertFalse(Z.changed('b'))

        Z.mark('a', False)
        Z.mark('b', True)

        self.assertTrue(A.changed())
        self.assertFalse(A.changed('x'))
        self.assertTrue(A.changed('y'))
        self.assertFalse(A.changed('z.a'))
        self.assertTrue(A.changed('z.b'))
        self.assertFalse(Z.changed('a'))
        self.assertTrue(Z.changed('b'))

    def testBitSetSubStruct(self):
        A = Value(Type([
            ('x', 'i'),
            ('y', 'i'),
            ('z', ('S', None, [
                ('a', 'i'),
                ('b', 'i'),
                ('q', ('S', None, [
                    ('m', 'i'),
                ])),
            ])),
        ]), {
        })

        self.assertSetEqual(A.changedSet(), set())
        A.mark('z')
        #self.assertSetEqual(A.changedSet(expand=False, parents=False), {'z'})
        self.assertSetEqual(A.changedSet(expand=True, parents=False),  {'z', 'z.a', 'z.b', 'z.q', 'z.q.m'})
        #self.assertSetEqual(A.changedSet(expand=False, parents=True),  {'z'})
        self.assertSetEqual(A.changedSet(expand=True, parents=True),   {'z', 'z.a', 'z.b', 'z.q', 'z.q.m'})

        A.unmark()
        A.mark('z.a')
        #self.assertSetEqual(A.changedSet(expand=False, parents=False), {     'z.a'})
        self.assertSetEqual(A.changedSet(expand=True, parents=False),  {     'z.a'})
        #self.assertSetEqual(A.changedSet(expand=False, parents=True),  {'z', 'z.a'})
        self.assertSetEqual(A.changedSet(expand=True, parents=True),   {'z', 'z.a'})

        A.unmark()
        A.mark('z.q')
        #self.assertSetEqual(A.changedSet(expand=False, parents=False), {     'z.q'})
        self.assertSetEqual(A.changedSet(expand=True, parents=False),  {     'z.q', 'z.q.m'})
        #self.assertSetEqual(A.changedSet(expand=False, parents=True),  {'z', 'z.q'})
        self.assertSetEqual(A.changedSet(expand=True, parents=True),   {'z', 'z.q', 'z.q.m'})

    def testTopAssignment(self):
        A = Value(Type([
            ('x', 'i'),
            ('y', 'i'),
            ('z', ('S', None, [
                ('a', 'i'),
                ('b', 'i'),
                ('c', 'i'),
            ])),
            ('q', 'v'),
        ]))

        A[None] = {'x':1, 'z':{'a':2}, 'z.b':3}

        self.assertEqual(A.x, 1)
        self.assertEqual(A.y, 0) # not changed
        self.assertEqual(A.z.a, 2)
        self.assertEqual(A.z.b, 3)
        self.assertSetEqual(A.changedSet(), {'x', 'z.a', 'z.b'})

    def testSubStructAssignment(self):
        A = Value(Type([
            ('x', 'i'),
            ('y', 'i'),
            ('z', ('S', None, [
                ('a', 'i'),
                ('b', 'i'),
                ('c', 'i'),
            ])),
            ('q', 'v'),
        ]), {
        })

        B = Value(Type([
            ('a', 'i'),
            ('b', 'i'),
            ('other', 'i'), # never marked, so not access during copy
        ]), {
            'a':42,
            'b':43,
        })

        A.z = B # copy alike structure (absense of 'c' is ok)
        A.q = B # variant union assignment (absense of 'c' also ok)

        self.assertEqual(A.z.a, 42)
        self.assertEqual(A.z.b, 43)
        self.assertEqual(A.q.a, 42)
        self.assertEqual(A.q.b, 43)
        self.assertSetEqual(A.changedSet(), {'z.a', 'z.b', 'q'})

class TestReInit(RefTestCase):

    def testCopySubStruct(self):
        A = Value(Type([
            ('x', ('S', None, [
                ('y', 'i'),
                ('z', 'ai'),
                ('q', 'v'),
                ('m', 'av'),
                ('a', ('S', None, [
                    ('A', 'i'),
                ])),
            ])),
        ]), {
            'x.y': 42,
            'x.z': range(3),
            'x.q': 'hello',
            'x.m': ['hello', 52],
            'x.a.A': 100,
        })

        B = Value(A.type(), {
            'x.y': 43,
            'x.z': range(4),
            'x.q': 15,
            'x.m': ['world', 62],
            'x.a.A': 101,
        })

        B.x = A.x

        self.assertEqual(B.x.y, 42)
        assert_aequal(B.x.z, [0, 1, 2])
        self.assertEqual(B.x.q, 'hello')
        self.assertEqual(B.x.m, ['hello', 52])
        self.assertEqual(B.x.a.A, 100)



# distutils: language = c++
#cython: language_level=2

cimport cython

from cpython.object cimport PyObject, PyTypeObject, traverseproc, visitproc
from cpython.ref cimport Py_INCREF, Py_XDECREF, Py_CLEAR

from libcpp cimport bool
from libcpp.cast cimport dynamic_cast
from libcpp.memory cimport unique_ptr, shared_ptr, weak_ptr
from libcpp.string cimport string
from libcpp.list cimport list as listxx
from libcpp.set cimport set as setxx
from libcpp.vector cimport vector

from .pvxs.client cimport Context, Report, ReportInfo
from .pvxs.server cimport ServerGUID
from .pvxs.source cimport Source, ChannelControl, OpBase, ClientCredentials
from . cimport _p4p

cdef extern from "pvxs_gw.h" namespace "p4p" nogil:
    enum: GWSearchIgnore
    enum: GWSearchClaim
    enum: GWSearchBanHost
    enum: GWSearchBanPV
    enum: GWSearchBanHostPV

    cdef cppclass GWChanInfo(ReportInfo):
        string usname

    ctypedef const GWChanInfo* GWChanInfoCP

    cdef cppclass GWUpstream:
        double get_holdoff

    cdef cppclass GWChan:
        const shared_ptr[GWUpstream] us
        const shared_ptr[ChannelControl] dschannel
        bool allow_put
        bool allow_rpc
        bool allow_uncached
        bool audit

    cdef cppclass GWSource(Source):
        Context upstream
        PyObject* handler

        @staticmethod
        shared_ptr[GWSource] build(const Context&) except+

        int test(const string&) except+
        shared_ptr[GWChan] connect(const string &dsname, const string &usname, unique_ptr[ChannelControl]* op) except+

        void sweep() except+
        void forceBan(const string& host, const string& usname) except+
        void clearBan() except+
        void cachePeek(setxx[string]& names) except+

        shared_ptr[GWSource] shared_from_this() except+

    shared_ptr[Source] makeOdometer(const string& name) except+

def addOdometer(_p4p.Server serv, basestring pvname, int order):
    cdef string name = pvname.encode()
    with nogil:
        serv.serv.addSource(name, makeOdometer(name), order)

cdef class InfoBase(object):
    cdef shared_ptr[const ClientCredentials] info

    @property
    def peer(self):
        if <bool>self.info:
            return self.info.get().peer.decode('UTF-8')
        else:
            return u''

    @property
    def account(self):
        if <bool>self.info:
            return self.info.get().account.decode('UTF-8')
        else:
            return u''

    @property
    def roles(self):
        cdef setxx[string] raw
        cdef list ret = []
        if <bool>self.info:
            for role in self.info.get().roles():
                ret.append(role.decode('UTF-8'))
        return ret

cdef class CreateOp(InfoBase):
    """Handle for in-progress Channel creation request
    """
    cdef readonly bytes name
    cdef unique_ptr[ChannelControl]* op
    cdef weak_ptr[GWSource] provider
    cdef object __weakref__

    def create(self, bytes name):
        """Create a Channel with a given upstream (server-side) name

        :param bytes name: Upstream name to use.  This is what the GW Client will search for.
        :returns: A `Channel`
        """
        cdef shared_ptr[GWChan] gwchan
        cdef Channel chan = Channel.__new__(Channel)
        cdef string dsname = self.name
        cdef string usname = name

        provider = self.provider.lock()
        if <bool>self.op and <bool>provider:
            with nogil:
                gwchan = provider.get().connect(dsname, usname, self.op)

            if not gwchan:
                raise RuntimeError("GW Provider will not create %s -> %s"%(usname, dsname))

            chan.channel = gwchan
            chan.name = name
            chan.info = gwchan.get().dschannel.get().credentials()
            return chan
        else:
            raise RuntimeError("Dead CreateOp")
            

cdef class Channel(InfoBase):
    cdef readonly bytes name
    cdef shared_ptr[GWChan] channel
    cdef object __weakref__

    @property
    def expired(self):
        return self.channel.use_count()<=1

    def access(self, put=None, rpc=None, uncached=None, audit=None, holdoff=None):
        if put is not None:
            self.channel.get().allow_put = put==True
        if rpc is not None:
            self.channel.get().allow_rpc = rpc==True
        if uncached is not None:
            self.channel.get().allow_uncached = uncached==True
        if audit is not None:
            self.channel.get().audit = audit==True
        if holdoff is not None:
            self.channel.get().us.get().get_holdoff = holdoff

@cython.no_gc_clear
cdef class Provider(_p4p.Source):
    cdef shared_ptr[GWSource] provider
    cdef object __weakref__
    cdef object dummy # ensure that this type participates in GC

    cdef readonly int Claim
    cdef readonly int Ignore
    cdef readonly int BanHost
    cdef readonly int BanPV
    cdef readonly int BanHostPV

    def __cinit__(self):
        self.Claim = GWSearchClaim
        self.Ignore = GWSearchIgnore
        self.BanHost = GWSearchBanHost
        self.BanPV = GWSearchBanPV
        self.BanHostPV = GWSearchBanHostPV

    def __init__(self, unicode name, object client, object handler):
        cdef _p4p.ClientProvider prov = client._ctxt
        cdef string cname = name.encode('utf-8')
        self.name = cname

        if not prov:
            raise ValueError('Not a Context')
        with nogil:
            self.provider = GWSource.build(prov.ctxt)
            self.src = <shared_ptr[Source]>self.provider

        Py_INCREF(handler)
        self.provider.get().handler = <PyObject*>handler

    def __dealloc__(self):
        if <bool>self.provider:
            Py_CLEAR(self.provider.get().handler)
        with nogil:
            self.provider.reset()

    def testChannel(self, bytes usname):
        """testChannel(usname)
        Add the upstream name to the channel cache and begin trying to connect.
        Returns Claim if the channel is connected, and Ignore if it is not.

        :param bytes usname: Upstream (Server side) PV name
        :returns: Claim or Ignore
        """
        cdef string n = usname
        cdef int ret
        with nogil:
            ret = self.provider.get().test(n)
        return ret

    def sweep(self):
        """Call periodically to remove unused `Channel` from channel cache.
        """
        with nogil:
            self.provider.get().sweep()

    def forceBan(self, bytes host = None, bytes usname = None):
        """Preemptively Add an entry to the negative result cache.
        Either host or usname must be not None

        :param bytes host: None or a host name
        :param bytes usname: None or a upstream (Server side) PV name
        """
        cdef string h
        cdef string us
        if host:
            h = host
        if usname:
            us = usname
        with nogil:
            self.provider.get().forceBan(h, us)

    def clearBan(self):
        """Clear the negative results cache
        """
        with nogil:
            self.provider.get().clearBan()

    def ignoreByGUID(self, list servers):
        cdef _p4p.Server serv
        cdef vector[ServerGUID] guids

        for ent in servers:
            serv = <_p4p.Server?>ent
            guids.push_back(serv.serv.config().guid)

        with nogil:
            self.provider.get().upstream.ignoreServerGUIDs(guids)

    def cachePeek(self):
        """Returns PV names in channel cache

        :returns: a set of strings
        """
        cdef setxx[string] pvs
        cdef set ret

        with nogil:
            self.provider.get().cachePeek(pvs)
        ret = set()
        for name in pvs:
            ret.add(name)
        return ret

    def stats(self):
        """Return statistics of various internal caches

        :rtype: dict
        """
        # TODO
        return {
            'ccacheSize.value':0,
            'mcacheSize.value':0,
            'gcacheSize.value':0,
            'banHostSize.value':0,
            'banPVSize.value':0,
            'banHostPVSize.value':0,
        }

    def report(self, float norm=1.0):
        """Run Client/Upstream bandwidth usage report

        :returns: List of tuple
        :rtype: [(usname, opTx, opRx, peer, trTx, trRx)]
        """
        cdef Report report
        cdef double cnorm = norm
        with nogil:
            report = self.provider.get().upstream.report()

        usinfo = []
        for conn in report.connections:
            peer = conn.peer.decode('utf-8', 'replace')
            for chan in conn.channels:
                usinfo.append((chan.name.decode('utf-8', 'replace'), chan.tx/cnorm, chan.rx/cnorm, peer, conn.tx/cnorm, conn.rx/cnorm))

        return usinfo

    def use_count(self):
        return self.provider.use_count()

def Server_report(_p4p.Server serv, float norm=1.0):
    """Return Server/Downstream bandwidth usage report

    :returns: List of tuple
    :rtype: [(usname, dsname, opTx, opRx, account, peer, trTx, trRx)]
    """
    cdef GWChanInfoCP info
    cdef Report report
    cdef double cnorm = norm
    with nogil:
        report = serv.serv.report()

    dsinfo = []
    for conn in report.connections:
        peer = conn.peer.decode('utf-8', 'replace')
        if <bool>conn.credentials:
            account = conn.credentials.get().account.decode('utf-8', 'replace')
        else:
            account = ''

        for chan in conn.channels:
            info = dynamic_cast[GWChanInfoCP](chan.info.get())

            if info:
                usname = info.usname.decode('utf-8', 'replace')
            else:
                usname = ''

            dsinfo.append((usname, chan.name.decode('utf-8', 'replace'), chan.tx/cnorm, chan.rx/cnorm, account, peer, conn.tx/cnorm, conn.rx/cnorm))

    return dsinfo

# Allow GC to find handler stored in GWProvider
#   https://github.com/cython/cython/issues/2737
cdef traverseproc Provider_base_traverse

cdef int holder_traverse(PyObject* raw, visitproc visit, void* arg) except -1:
    cdef int ret = 0
    cdef Provider self = <Provider>raw

    if self.provider.get().handler:
        visit(self.provider.get().handler, arg)
    return Provider_base_traverse(raw, visit, arg)

Provider_base_traverse = (<PyTypeObject*>Provider).tp_traverse
(<PyTypeObject*>Provider).tp_traverse = <traverseproc>holder_traverse

cdef public:
    int GWProvider_testChannel(object handler, const char* channel, const char* peer) with gil:
        if not handler:
            return GWSearchBanHost
        try:
            return handler.testChannel(channel, peer.decode())
        except:
            return GWSearchBanHost

    shared_ptr[GWChan] GWProvider_makeChannel(GWSource* src, unique_ptr[ChannelControl]* op) with gil:
        cdef shared_ptr[GWChan] ret
        cdef CreateOp create
        if not src.handler:
            return ret

        handler = <object>src.handler

        create = CreateOp.__new__(CreateOp)
        create.name = op.get().name().c_str()
        create.op = op
        create.info = op.get().credentials()
        create.provider = <weak_ptr[GWSource]>src.shared_from_this()

        chan = handler.makeChannel(create)

        create.op = NULL # makeChannel() must decide immediately

        if chan is not None:
            ret = (<Channel?>chan).channel

        return ret

    void GWProvider_audit(GWSource* src, listxx[string]& cmsgs) with gil:
        if src.handler:
            handler = <object>src.handler
            msgs = [msg.decode() for msg in cmsgs]
            try:
                handler.audit(msgs)
            except:
                pass # expect handler to catch and log



import sys
import unittest

try:
    from qtpy.QtCore import QObject, QCoreApplication
except ImportError as e:
    class TestDummy(unittest.TestCase):
        def test_asyncio(self):
            raise unittest.SkipTest("Missing qtpy")

else:
    from .qttest import *



#cython: language_level=2

from libc.stdint cimport uint8_t
from libcpp.string cimport string
from libcpp.map cimport map

cdef extern from "<pvxs/util.h>" namespace "pvxs" nogil:
    map[string, size_t] instanceSnapshot() except+

    cdef cppclass ServerGUID: # really derivies from std::array
        size_t size() const
        uint8_t& operator[](size_t)
        uint8_t* data()




from __future__ import print_function
"""Python reference counter statistics."""

import sys
import gc
import inspect
import time
from glob import fnmatch
try:
    from types import InstanceType
except ImportError:  # py3
    InstanceType = None


class StatsDelta(object):

    """GC statistics tracking.

    Monitors the number of instances of each type/class (cf. gcstats())
    and prints a report of any changes (ie. new types/count changes).
    Intended to assist in detecting reference leaks.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        """Reset internal statistics counters
        """
        self.stats, self.ntypes = None, None

    def collect(self, file=sys.stderr):
        """Collect stats and print results to file

        :param file: A writable file-like object
        """
        cur = gcstats()
        Ncur = len(cur)
        if self.stats is not None and file is not None:
            prev = self.stats
            Nprev = self.ntypes  # may be less than len(prev)

            if Ncur != Nprev:
                print("# Types %d -> %d" % (Nprev, Ncur), file=file)

            Scur, Sprev, first = set(cur), set(prev), True
            for T in Scur - Sprev:  # new types
                if first:
                    print('New Types', file=file)
                    first = False
                print(' ', T, cur[T], file=file)

            first = True
            for T in Sprev - Scur:  # collected types
                if first:
                    print('Cleaned Types', file=file)
                    first = False
                print(' ', T, -prev[T], file=file)

            first = True
            for T in Scur & Sprev:
                if cur[T] == prev[T]:
                    continue
                if first:
                    print('Known Types', file=file)
                    first = False
                print(' ', T, cur[T], 'delta', cur[T] - prev[T], file=file)

        else:  # first call
            print("All Types", file=file)
            for T, C in cur.items():
                print(' ', T, C, file=file)

        self.stats, self.ntypes = cur, len(cur)
        # gc.collect()


def gcstats():
    """Count the number of instances of each type/class

    :returns: A dict() mapping type (as a string) to an integer number of references
    """
    all = gc.get_objects()
    _stats = {}

    for obj in all:
        K = type(obj)
        if K is StatsDelta:
            continue  # avoid counting ourselves

        elif K is InstanceType:  # instance of an old-style class
            K = getattr(obj, '__class__', K)

        # Track types as strings to avoid holding references
        K = str(K)

        try:
            _stats[K] += 1
        except KeyError:
            _stats[K] = 1

    # explicitly break the reference loop between the list and this frame,
    # which is contained in the list
    # This would otherwise prevent the list from being free'd
    del all

    return _stats


def gcfind(name):
    all = gc.get_objects()
    found = []

    for obj in all:
        K = type(obj)
        if K is gcfind:
            continue  # avoid counting ourselves

        if K is InstanceType:  # instance of an old-style class
            K = getattr(obj, '__class__', K)

        if fnmatch(str(K), name):
            found.append(obj)

    return found


class _StatsThread(object):

    def __init__(self, period, file):
        self.period, self.file = period, file
        self.S = StatsDelta()

    def __call__(self):
        while True:
            self.S.collect(file=self.file)
            time.sleep(self.period)


def periodic(period=60.0, file=sys.stderr):
    """Start a daemon thread which will periodically print GC stats

    :param period: Update period in seconds
    :param file: A writable file-like object
    """
    import threading
    import time
    S = _StatsThread(period=period, file=file)
    T = threading.Thread(target=S)
    T.daemon = True
    T.start()

if __name__ == '__main__':
    # for T,C in gcstats().items():
    #  print T,C
    gc.set_debug(gc.DEBUG_COLLECTABLE)
    S = StatsDelta()
    while True:
        S.collect()
        # gc.collect()




import logging
_log = logging.getLogger(__name__)

from functools import partial

try:
    from Queue import Queue, Full, Empty
except ImportError:
    from queue import Queue, Full, Empty
from threading import Thread, Event

__all__ = [
    'WorkQueue',
]


class WorkQueue(object):

    """A threaded work queue.
    """
    _stopit = object()

    def __init__(self, maxsize=5):
        self._Q = Queue(maxsize=maxsize)

    def push(self, callable):
        self._Q.put_nowait(callable)  # throws Queue.Full

    def push_wait(self, callable):
        self._Q.put(callable)

    def interrupt(self):
        """Break one call to handle()

        eg. Call N times to break N threads.

        This call blocks if the queue is full.
        """
        self._Q.put(self._stopit)

    def handle(self):
        """Process queued work until interrupt() is called
        """
        while True:
            # TODO: Queue.get() (and anything using thread.allocate_lock
            #       ignores signals :(  so timeout periodically to allow delivery
            try:
                callable = None # ensure no lingering references to past work while blocking
                callable = self._Q.get(True, 1.0)
            except Empty:
                continue  # retry on timeout
            try:
                if callable is self._stopit:
                    break
                callable()
            except:
                _log.exception("Error from WorkQueue w/ %r", callable)
            finally:
                self._Q.task_done()

class ThreadedWorkQueue(WorkQueue):
    def __init__(self, name=None, workers=1, daemon=False, **kws):
        assert workers>=1, workers
        WorkQueue.__init__(self, **kws)
        self.name = name
        self._daemon = daemon
        self._T = [None]*workers

    def __enter__(self):
        self.start()
    def __exit__(self, A,B,C):
        self.stop()

    def start(self):
        for n in range(len(self._T)):
            if self._T[n] is not None:
                continue
            T = self._T[n] = Thread(name='%s[%d]'%(self.name, n), target=self.handle)
            T.daemon = self._daemon
            T.start()

        return self # allow chaining

    def stop(self):
        [self.interrupt() for T in self._T if T is not None]
        [T.join()      for T in self._T if T is not None]
        self._T = [None]*len(self._T)

        return self # allow chaining

    def sync(self, timeout=None):
        wait1 = [Event() for _n in range(len(self._T))]
        wait2 = [Event() for _n in range(len(self._T))]

        def syncX(wait1, wait2):
            wait1.set()
            wait2.wait()

        [self.push_wait(partial(syncX, wait1[n], wait2[n])) for n in range(len(self._T))]

        # wait for all workers to ready wait1 barrier
        for W in wait1:
            W.wait(timeout=timeout)

        # allow workers to proceeed
        for W in wait2:
            W.set()

        return self # allow chaining

# lazy create a default work queues

class _DefaultWorkQueue(object):

    def __init__(self, workers=4):  # TODO: configurable?
        self.W = [None]*workers
        self.n = 0

    def __del__(self):
        self.stop()

    def __call__(self):
        W = self.W[self.n]
        if W is None:
            #  daemon=True  otherwise the MainThread exit handler tries to join too early
            W = self.W[self.n] = ThreadedWorkQueue(maxsize=0, daemon=True).start()

        # sort of load balancing by giving different queues to each SharedPV
        # but preserve ordering or callbacks as each SharedPV has only one queue
        self.n = (self.n+1)%len(self.W)
        return W

    def sync(self):
        [W.sync() for W in self.W if W is not None]

    def stop(self):
        [W.stop() for W in self.W if W is not None]
        self.W = [None]*len(self.W)

_defaultWorkQueue = _DefaultWorkQueue()



#cython: language_level=2

from libcpp cimport bool
from libcpp.string cimport string
from libcpp.map cimport map
from libcpp.list cimport list as clist
from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr
#from libcpp.functional cimport function

from .data cimport Value
from .util cimport ServerGUID
from .source cimport ClientCredentials

cdef extern from "<pvxs/client.h>" namespace "pvxs::client" nogil:
    ctypedef map[string, string] Config_defs_t "pvxs::client::Config::defs_t"

    cdef cppclass Config:
        vector[string] addressList
        vector[string] interfaces
        unsigned short udp_port
        bool autoAddrList

        @staticmethod
        Config fromEnv() except+
        @staticmethod
        Config fromDefs(const map[string, string]&) except+

        Config& applyEnv() except+
        Config& applyDefs(const Config_defs_t&) except+
        void updateDefs(Config_defs_t& defs) except+

        void expand() except+

        Context build() except+

    enum cacheAction "pvxs::client::Context::cacheAction":
        Clean "pvxs::client::Context::Clean"
        Drop "pvxs::client::Context::Drop"
        Disconnect "pvxs::client::Context::Disconnect"

    cdef cppclass Context:
        Context()
        Context(const Config&) except+

        # methods returning builds must be except+ to workaround
        # https://github.com/cython/cython/issues/3759

        Config config() except+

        GetBuilder get(const string& pvname) except+
        PutBuilder put(const string& pvname) except+
        RPCBuilder rpc(const string& pvname) except+
        RPCBuilder rpc(const string& pvname, const Value&) except+
        MonitorBuilder monitor(const string& pvname) except+

        @staticmethod
        RequestBuilder request() except+

        void hurryUp() except+
        void cacheClear(const string& name, cacheAction action) except+
        void ignoreServerGUIDs(const vector[ServerGUID]& guids) except+

        Report report() except+

    cdef cppclass GetBuilder:
        GetBuilder& field(const string& fld) except+
        GetBuilder& record[T](const string& name, const T& val) except+
        GetBuilder& pvRequest(const string& fld) except+
        GetBuilder& rawRequest(const Value& fld) except+

        #GetBuilder& result(function[void(Result&&)]&& cb) except+

        shared_ptr[Operation] exec_ "exec"() except+

    cdef cppclass PutBuilder:
        PutBuilder& field(const string& fld) except+
        PutBuilder& record[T](const string& name, const T& val) except+
        PutBuilder& pvRequest(const string& fld) except+
        PutBuilder& rawRequest(const Value& fld) except+

        PutBuilder& fetchPresent(bool f)
        #PutBuilder& build(std::function[Value(Value&&)]&& cb)
        #PutBuilder& result(function[void(Result&&)]&& cb) except+

        shared_ptr[Operation] exec_ "exec"() except+

    cdef cppclass RPCBuilder:
        RPCBuilder& field(const string& fld) except+
        RPCBuilder& record[T](const string& name, const T& val) except+
        RPCBuilder& pvRequest(const string& fld) except+
        RPCBuilder& rawRequest(const Value& fld) except+

        #RPCBuilder& result(function[void(Result&&)]&& cb) except+

        shared_ptr[Operation] exec_ "exec"() except+

    cdef cppclass MonitorBuilder:
        MonitorBuilder& field(const string& fld) except+
        MonitorBuilder& record[T](const string& name, const T& val) except+
        MonitorBuilder& pvRequest(const string& fld) except+
        MonitorBuilder& rawRequest(const Value& fld) except+

        #MonitorBuilder& event(function[void(Subscription&)]&& cb)
        MonitorBuilder& maskConnected(bool m)
        MonitorBuilder& maskDisconnected(bool m)

        shared_ptr[Subscription] exec_ "exec"() except+

    cdef cppclass RequestBuilder:
        RequestBuilder& field(const string& fld) except+
        RequestBuilder& record[T](const string& name, const T& val) except+
        RequestBuilder& pvRequest(const string& fld) except+

        Value build() except+

    cdef cppclass Subscription:
        bool cancel() except+
        void pause(bool p) except+
        void resume() except+
        Value pop() except+

    cdef cppclass Result:
        pass

    cdef cppclass Operation:
        bool cancel() except+


    # really netcommon.h, but can't include this directly
    cdef cppclass ReportInfo "pvxs::client::ReportInfo":
        pass

    cdef cppclass Channel "pvxs::client::Report::Channel":
        string name
        size_t tx
        size_t rx
        shared_ptr[const ReportInfo] info

    cdef cppclass Connection "pvxs::client::Report::Connection":
        string peer
        shared_ptr[ClientCredentials] credentials
        size_t tx
        size_t rx
        clist[Channel] channels

    cdef cppclass Report:
        clist[Connection] connections



import logging
import unittest
from functools import wraps
from math import ceil

from qtpy.QtCore import QObject, QCoreApplication

from ..client.Qt import exceptionGuard, Disconnected, TimeoutError, Context

from ..server import Server, StaticProvider
from ..server.thread import SharedPV, _defaultWorkQueue
from ..nt import NTScalar
from .utils import RefTestCase

_log = logging.getLogger(__name__)

class waitSignal(QObject):
    def __init__(self, sig=None):
        QObject.__init__(self)
        self._T = None
        self._args = None
        self._sig = sig

    def __enter__(self):
        if self._sig is not None:
            self._sig.connect(self.slot)
        return self

    def __exit__(self,A,B,C):
        self.close()
        if self._sig is not None:
            self._sig.disconnect(self.slot)

    def close(self):
        if self._T is not None:
            self.killTimer(self._T)
            self._T = None

    @exceptionGuard
    def timerEvent(self, evt):
        _log.error('Test timeout')
        QCoreApplication.instance().quit()
        self.close()

    def slot(self, *args):
        _log.debug('signalled %s', args)
        self._args = args
        QCoreApplication.instance().quit()
        self.close()

    def wait(self, timeout=5.0):
        assert self._T is None, self._T
        self._T = self.startTimer(ceil(timeout*1000))
        try:
            self._args = None
            QCoreApplication.instance().exec_()
            if self._args is not None:
                return self._args

            else:
                raise TimeoutError()
        finally:
            self.close()

    def waitValue(self, **kws):
        E, = self.wait(**kws)
        if isinstance(E, Exception):
            raise E
        else:
            return E

class TestGPM(RefTestCase):
    def setUp(self):
        super(TestGPM, self).setUp()

        self._timedout = False
        self.app = QCoreApplication([])

        self.pv = SharedPV(nt=NTScalar('i'), initial=42)
        @self.pv.put
        def mailbox(pv, op):
            _log.debug('onPut: %s', op.value())
            pv.post(op.value())
            op.done()
        self.server = Server(providers=[{'pvname':self.pv}], isolate=True)

        self.ctxt = Context('pva', conf=self.server.conf(), useenv=False)

    def tearDown(self):
        self.ctxt.close()

        self.server.stop()
        _defaultWorkQueue.sync()
        del self.ctxt
        del self.server
        del self.pv
        del self.app
        super(TestGPM, self).tearDown()

    def test_pm(self):
        with waitSignal() as sub, waitSignal() as put:
            S = self.ctxt.monitor('pvname', sub.slot)
            E, = sub.wait()
            self.assertIsInstance(E, Disconnected)

            self.assertEqual(sub.waitValue(), 42)

            self.ctxt.put('pvname', 51, slot=put.slot)
            put.waitValue()

            self.assertEqual(sub.waitValue(), 51)




import time
from datetime import datetime
from ..wrapper import Type

if hasattr(datetime, 'timestamp'):
    _dt2posix = datetime.timestamp

else: # py 2.x
    from calendar import timegm as _timegm
    def _dt2posix(dt):
        return _timegm(dt.timetuple()) * dt.microsecond*1e-6

# common sub-structs
timeStamp = Type(id='time_t', spec=[
    ('secondsPastEpoch', 'l'),
    ('nanoseconds', 'i'),
    ('userTag', 'i'),
])
alarm = Type(id='alarm_t', spec=[
    ('severity', 'i'),
    ('status', 'i'),
    ('message', 's'),
])

class NTBase(object):
    """Helper for NT* (un)wrapper classes
    """
    @staticmethod
    def buildType():
        """Return a Type
        """
        raise NotImplementedError("NT classes must provide buildType()")

    def __init__(self, *args, **kws):
        self.type = self.buildType(*args, **kws)

    def wrap(self, pyval, **kws):
        """Called during eg. SharedPV.post() to build a Value
        """
        raise NotImplementedError("NT classes must provide wrap()")

    def assign(self, V, pyval):
        """Called during eg. Context.put() to update the Value V from the provided python object
        """
        return self.wrap(self.buildType(), pyval)

    def unwrap(self, V):
        """Translate a Value into an appropriate python object.  (the reverse of wrap())
        """
        raise NotImplementedError("NT classes must provide unwrap()")

    @staticmethod
    def _annotate(value, timestamp=None, severity=None, message=None):
        """Update and return provided 'value'.

        Updates timeStamp.secondsPastEpoch, timeStamp.nanoseconds with
        the provided timestamp or the current system time.
        """
        if severity is not None:
            value['alarm.severity'] = severity

        if message is not None:
            value['alarm.message'] = message

        if timestamp is not None:
            # timestamp may be: datetime, seconds as float or int, or tuple of (sec, ns)
            if isinstance(timestamp, datetime):
                timestamp = _dt2posix(timestamp)

            if isinstance(timestamp, (int, float)):
                sec, ns = divmod(timestamp, 1.0)
                timestamp = (int(sec), int(ns*1e9))

            # at this point timestamp must be a tuple of (sec, ns)
            value['timeStamp'] = {'secondsPastEpoch':timestamp[0], 'nanoseconds':timestamp[1]}
        return value



#cython: language_level=2

cdef extern from "<pvxs/log.h>" namespace "pvxs" nogil:
    enum Level:
        Debug "pvxs::Level::Debug"
        Info "pvxs::Level::Info"
        Warn "pvxs::Level::Warn"
        Err "pvxs::Level::Err"
        Crit "pvxs::Level::Crit"

    void logger_level_set(const char *name, int lvl) except+
    void logger_level_clear() except+
    void logger_config_env() except+




from __future__ import print_function

import logging
_log = logging.getLogger(__name__)

import warnings
import sys

try:
    from Queue import Queue, Full, Empty
except ImportError:
    from queue import Queue, Full, Empty

from .. import _p4p
from .._p4p import Cancelled, Disconnected, Finished, RemoteError

from ..wrapper import Value, Type
from ..nt import buildNT

if sys.version_info >= (3, 0):
    unicode = str

__all__ = (
    'Subscription',
    'Context',
    'RemoteError',
)


def unwrapHandler(handler, nt):
    """Wrap get/rpc handler to unwrap Value
    """
    def dounwrap(code, msg, val, handler=handler):
        _log.debug("Handler (%s, %s, %r) -> %s", code, msg, val, handler)
        try:
            if code == 0:
                handler(RemoteError(msg))
            elif code == 1:
                handler(Cancelled())
            elif code == 2: # exception during builder callback
                A, B, C = val
                if unicode is str:
                    E = A(B).with_traceback(C) # py 3
                else:
                    E = A(B) # py 2 (bye bye traceback...)
                handler(E)
            else:
                if val is not None:
                    val = nt.unwrap(val)
                handler(val)
        except:
            _log.exception("Exception in Operation handler")
    return dounwrap


def monHandler(handler):
    def cb(handler=handler):
        _log.debug("Update %s", handler)
        try:
            handler()
        except:
            _log.exception("Exception in Monitor handler")
    return cb


def defaultBuilder(value, nt):
    """Reasonably sensible default handling of put builder
    """
    if callable(value):
        return value

    def builder(V):
        if isinstance(value, Value):
            V[None] = value
        elif isinstance(value, dict):
            for k, v in value.items():
                V[k] = v
        else:
            nt.assign(V, value)
    return builder


def wrapRequest(request):
    if request is None or isinstance(request, Value):
        return request
    return Context.makeRequest(request)


class Subscription(_p4p.ClientMonitor):

    """Interface to monitor subscription FIFO

    Use method poll() to try to pop an item from the FIFO.
    None indicates the FIFO is empty, must wait for another Data event before
    calling poll() again.

    complete()==True after poll()==False indicates that no more updates will
    ever be forthcoming.  This is normal (not error) completion.

    cancel() aborts the subscription.
    """

    def __init__(self, context, name, nt, **kws):
        _log.debug("Subscription(%s)", kws)
        super(Subscription, self).__init__(context, name, **kws)
        self.context = context
        self._nt = nt
        self.done = False

    def pop(self):
        val = super(Subscription, self).pop()
        assert val is None or isinstance(val, (Value, Exception)), val
        if isinstance(val, Value):
            val = self._nt.unwrap(val)
        elif isinstance(val, Finished):
            self.done = True
        _log.debug("poll() -> %r", val)
        return val

    def complete(self):
        return self.done

    def __enter__(self):
        return self

    def __exit__(self, A, B, C):
        self.close()

    if unicode is str:
        def __del__(self):
            self.close()

class Context(object):

    """
    :param str provider: A Provider name.  Try "pva" or run :py:meth:`Context.providers` for a complete list.
    :param conf dict: Configuration to pass to provider.  Depends on provider selected.
    :param bool useenv: Allow the provider to use configuration from the process environment.
    :param dict nt: Controls :ref:`unwrap`.  None uses defaults.  Set False to disable
    :param dict unwrap: Legacy :ref:`unwrap`.
    """

    def __init__(self, provider='pva', conf=None, useenv=None,
                 unwrap=None, nt=None,
                 **kws):
        self.name = provider
        super(Context, self).__init__(**kws)

        self._nt = buildNT(nt, unwrap)

        self._ctxt = None

        self._ctxt = _ClientProvider(provider, conf=conf, useenv=useenv)
        self.conf = self._ctxt.conf
        self.hurryUp = self._ctxt.hurryUp

    makeRequest = _p4p.ClientProvider.makeRequest

    def close(self):
        if self._ctxt is None:
            return

        self._ctxt.close()
        self._ctxt = None

    def __enter__(self):
        return self

    def __exit__(self, A, B, C):
        self.close()

    def disconnect(self, name=None):
        """Clear internal Channel cache, allowing currently unused channels to be implictly closed.

        :param str name: None, to clear the entire cache, or a name string to clear only a certain entry.
        """
        self._ctxt.disconnect(name)

    def _request(self, process=None, wait=None):
        """helper for building pvRequests

        :param str process: Control remote processing.  May be 'true', 'false', 'passive', or None.
        :param bool wait: Wait for all server processing to complete.
        """
        opts = []
        if process is not None:
            opts.append('process=%s' % process)
        if wait is not None:
            if wait:
                opts.append('wait=true')
            else:
                opts.append('wait=false')
        return 'field()record[%s]' % (','.join(opts))

    def get(self, name, handler, request=None):
        """Begin Fetch of current value of a PV

        :param name: A single name string or list of name strings
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param callable handler: Completion notification.  Called with a Value, RemoteError, or Cancelled

        :returns: A object with a method cancel() which may be used to abort the operation.
        """
        return _ClientOperation(self._ctxt, name, handler=unwrapHandler(handler, self._nt),
                                    pvRequest=wrapRequest(request), get=True, put=False)

    def put(self, name, handler, builder=None, request=None, get=True):
        """Write a new value to a PV.

        :param name: A single name string or list of name strings
        :param callable handler: Completion notification.  Called with None (success), RemoteError, or Cancelled
        :param callable builder: Called when the PV Put type is known.  A builder is responsible
                                 for filling in the Value to be sent.  builder(value)
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param bool get: Whether to do a Get before the Put.  If True then the value passed to the builder callable
                         will be initialized with recent PV values.  eg. use this with NTEnum to find the enumeration list.

        :returns: A object with a method cancel() which may be used to abort the operation.
        """
        return _ClientOperation(self._ctxt, name, handler=unwrapHandler(handler, self._nt),
                                    builder=defaultBuilder(builder, self._nt),
                                    pvRequest=wrapRequest(request), get=get, put=True)

    def rpc(self, name, handler, value, request=None):
        """Perform RPC operation on PV

        :param name: A single name string or list of name strings
        :param callable handler: Completion notification.  Called with a Value, RemoteError, or Cancelled
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.

        :returns: A object with a method cancel() which may be used to abort the operation.
        """
        if value is None:
            value = Value(Type([]))
        return _ClientOperation(self._ctxt, name, handler=unwrapHandler(handler, self._nt),
                                    value=value, pvRequest=wrapRequest(request), rpc=True)

    def monitor(self, name, handler, request=None, **kws):
        """Begin subscription to named PV

        :param str name: PV name string
        :param callable handler: Completion notification.  Called with None (FIFO not empty), RemoteError, Cancelled, or Disconnected
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param bool notify_disconnect: Whether disconnect (and done) notifications are delivered to the callback (as None).

        :returns: A Subscription
        """
        return Subscription(self._ctxt, name,
                            nt=self._nt,
                            handler=monHandler(handler), pvRequest=wrapRequest(request),
                            **kws)

    @staticmethod
    def providers():
        return ["pva"]

    @staticmethod
    def set_debug(lvl):
        _p4p.set_debug(lvl)

set_debug = _p4p.logger_level_set

def _cleanup_contexts():
    contexts = list(_p4p.all_providers)
    _log.debug("Closing %d Client contexts", len(contexts))
    for ctxt in contexts:
        ctxt.close()

class _ClientOperation(_p4p.ClientOperation):
    if unicode is str:
        def __del__(self):
            self.close()

class _ClientProvider(_p4p.ClientProvider):
    if unicode is str:
        def __del__(self):
            self.close()




import sys
import unittest

try:
    import cothread
except ImportError:
    class TestDummy(unittest.TestCase):
        def test_asyncio(self):
            raise unittest.SkipTest("Missing cothread")

else:
    from .cothreadtest import *




from __future__ import print_function

import logging
import sys
import gc
import inspect
import unittest
import time
import os
import tempfile
import fnmatch
import weakref

from functools import wraps

from .. import listRefs
from .._p4p import _forceLazy

_log = logging.getLogger(__name__)

_forceLazy()

if not hasattr(unittest.TestCase, 'assertRegex'):
    unittest.TestCase.assertRegex = unittest.TestCase.assertRegexpMatches

if not hasattr(unittest.TestCase, 'assertRaisesRegex'):
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp

class RefTestMixin(object):
    __showLeftovers = True

    """Ensure that each test does not result in a net change in extension object counts
    """
    # set to list of names to compare.  Set to None to disable
    ref_check = ('*',)

    def __refs(self, refs=None):
        refs = refs or listRefs()
        _log.debug("REFS %s", refs)
        names = set()
        for pat in self.ref_check:
            names |= set(fnmatch.filter(refs, pat))
        return dict([(K, V) for K, V in refs.items() if K in names and V>0])

    def setUp(self):
        self.__traceme = set()
        if self.ref_check is not None:
            self.__before = self.__refs()

            for mustzero in ('ClientContextImpl',):
                if self.__before.get(mustzero, 0)!=0 and self.__showLeftovers:
                    self.__showLeftovers = False # only show failure once
                    self.fail('Leftovers from previous test: %s = %d'%(mustzero, self.__before[mustzero]))

        super(RefTestMixin, self).setUp()

    def traceme(self, obj):
        self.__traceme.add(weakref.ref(obj))

    def _sleep(self, delay):
        time.sleep(delay)

    def tearDown(self):
        super(RefTestMixin, self).tearDown()
        if self.ref_check is not None:
            traceme = list(self.__traceme)
            del self.__traceme
            gc.collect()
            after = self.__refs()

            test = self.__before == after

            for mustzero in ('ClientContextImpl',):
                test &= after.get(mustzero, 0)==0

            frame = inspect.currentframe()
            for T in traceme:
                O = T()
                if O is None:
                    continue
                nrefs = sys.getrefcount(O)
                refs = gc.get_referrers(O)
                nrefs -= len(refs) # exclude tracked refs
                refs = filter(lambda o:o not in (frame, traceme), refs)
                _log.debug("ALIVE %s -> %s + %d ext refs", O, refs, nrefs)

            self.assertDictEqual(self.__before, after)
            # check for any obviously corrupt counters, even those not being compared
            # self.assertFalse(any([V>1000000 for V in refs.values()]), "before %s after %s"%(self.__raw_before, refs))

            if not test:
                for mustzero in ('ClientContextImpl', 'ServerPvt'):
                    self.assertEqual(0, after.get(mustzero, 0), mustzero)
                self.assertDictEqual(self.__before, after)

class RefTestCase(RefTestMixin, unittest.TestCase):
    def __init__(self, methodName='runTest'):
        # skip reference check for tests which have already failed.
        meth = getattr(self, methodName)
        @wraps(meth)
        def wrapper(*args, **kws):
            try:
                return meth(*args, **kws)
            except:
                self.ref_check = None
                raise
        setattr(self, methodName, wrapper)
        super(RefTestCase, self).__init__(methodName=methodName)

    def setUp(self):
        super(RefTestCase, self).setUp()

    def tearDown(self):
        super(RefTestCase, self).tearDown()

    if not hasattr(unittest.TestCase, 'assertRegex'):
        def assertRegex(self, text, regex):
            import re
            self.assertTrue(re.search(regex, text),
                            """Regex didn't match: %r not found in %r"""%(regex, text))


def gctrace(obj, maxdepth=8):
    # depth first traversal
    pop = object()
    top = inspect.currentframe()
    next = top.f_back
    stack, todo = [], [obj]
    visited = set()

    while len(todo):
        obj = todo.pop(0)
        I = id(obj)
        if inspect.isframe(obj):
            S = 'Frame %s:%d' % (obj.f_code.co_filename, obj.f_lineno)
        else:
            S = str(obj)

        if obj is pop:
            stack.pop()
            # break
            continue

        print('-' * len(stack), S, end='')

        if I in stack:
            print(' Recurse')
            continue
        elif I in visited:
            print(' Visited')
            continue
        elif len(stack) >= maxdepth:
            print(' Depth limit')
            continue
        else:
            print(' ->')

        stack.append(I)
        visited.add(I)

        todo.insert(0, pop)

        for R in gc.get_referrers(obj):
            if R is top or R is next or R is todo:
                continue
            todo.insert(0, R)

class RegularNamedTemporaryFile(object):
    """Like tempfile.NamedTemporaryFile which doesn't use O_TEMPORARY on windows
    """
    def __init__(self, *args, **kws):
        mkstemp_kws = {k: v for k, v in kws.items()
                       if k in ['suffix', 'prefix', 'dir']}
        fd, self.name = tempfile.mkstemp(**mkstemp_kws)
        try:
            if 'mode' in kws:
                self.file = os.fdopen(fd, kws['mode'])
            else:
                self.file = os.fdopen(fd)
            self.read = self.file.read
            self.write = self.file.write
            self.flush = self.file.flush
            self.seek = self.file.seek
        except:
            os.unlink(self.name)
            raise

    def __del__(self):
        self.close()

    def __enter__(self):
        return self
    def __exit__(self,A,B,C):
        self.close()

    def close(self):
        if self.file is not None:
            self.file.close()
            os.unlink(self.name)
            self.file = None
            self.read = self.write = self.flush = self.seek = None
